import unittest.mock
from unittest.mock import call

import pydantic
import pytest
from litellm.exceptions import ContextWindowExceededError

import strands
from strands.models.litellm import LiteLLMModel
from strands.types.exceptions import ContextWindowOverflowException


@pytest.fixture
def litellm_acompletion():
    with unittest.mock.patch.object(strands.models.litellm.litellm, "acompletion") as mock_acompletion:
        yield mock_acompletion


@pytest.fixture
def api_key():
    return "a1"


@pytest.fixture
def model_id():
    return "m1"


@pytest.fixture
def model(litellm_acompletion, api_key, model_id):
    _ = litellm_acompletion

    return LiteLLMModel(client_args={"api_key": api_key}, model_id=model_id)


@pytest.fixture
def messages():
    return [{"role": "user", "content": [{"text": "test"}]}]


@pytest.fixture
def system_prompt():
    return "s1"


@pytest.fixture
def test_output_model_cls():
    class TestOutputModel(pydantic.BaseModel):
        name: str
        age: int

    return TestOutputModel


def test_update_config(model, model_id):
    model.update_config(model_id=model_id)

    tru_model_id = model.get_config().get("model_id")
    exp_model_id = model_id

    assert tru_model_id == exp_model_id


@pytest.mark.parametrize(
    "client_args, model_id, expected_model_id",
    [
        ({"use_litellm_proxy": True}, "openai/gpt-4", "litellm_proxy/openai/gpt-4"),
        ({"use_litellm_proxy": False}, "openai/gpt-4", "openai/gpt-4"),
        ({"use_litellm_proxy": None}, "openai/gpt-4", "openai/gpt-4"),
        ({}, "openai/gpt-4", "openai/gpt-4"),
        (None, "openai/gpt-4", "openai/gpt-4"),
        ({"use_litellm_proxy": True}, "litellm_proxy/openai/gpt-4", "litellm_proxy/openai/gpt-4"),
        ({"use_litellm_proxy": False}, "litellm_proxy/openai/gpt-4", "litellm_proxy/openai/gpt-4"),
    ],
)
def test__init__use_litellm_proxy_prefix(client_args, model_id, expected_model_id):
    """Test litellm_proxy prefix behavior for various configurations."""
    model = LiteLLMModel(client_args=client_args, model_id=model_id)
    assert model.get_config()["model_id"] == expected_model_id


@pytest.mark.parametrize(
    "client_args, initial_model_id, new_model_id, expected_model_id",
    [
        ({"use_litellm_proxy": True}, "openai/gpt-4", "anthropic/claude-3", "litellm_proxy/anthropic/claude-3"),
        ({"use_litellm_proxy": False}, "openai/gpt-4", "anthropic/claude-3", "anthropic/claude-3"),
        (None, "openai/gpt-4", "anthropic/claude-3", "anthropic/claude-3"),
    ],
)
def test_update_config_proxy_prefix(client_args, initial_model_id, new_model_id, expected_model_id):
    """Test that update_config applies proxy prefix correctly."""
    model = LiteLLMModel(client_args=client_args, model_id=initial_model_id)
    model.update_config(model_id=new_model_id)
    assert model.get_config()["model_id"] == expected_model_id


@pytest.mark.parametrize(
    "content, exp_result",
    [
        # Case 1: Thinking
        (
            {
                "reasoningContent": {
                    "reasoningText": {
                        "signature": "reasoning_signature",
                        "text": "reasoning_text",
                    },
                },
            },
            {
                "signature": "reasoning_signature",
                "thinking": "reasoning_text",
                "type": "thinking",
            },
        ),
        # Case 2: Video
        (
            {
                "video": {
                    "source": {"bytes": "base64encodedvideo"},
                },
            },
            {
                "type": "video_url",
                "video_url": {
                    "detail": "auto",
                    "url": "base64encodedvideo",
                },
            },
        ),
        # Case 3: Text
        (
            {"text": "hello"},
            {"type": "text", "text": "hello"},
        ),
    ],
)
def test_format_request_message_content(content, exp_result):
    tru_result = LiteLLMModel.format_request_message_content(content)
    assert tru_result == exp_result


@pytest.mark.asyncio
async def test_stream(litellm_acompletion, api_key, model_id, model, agenerator, alist):
    mock_delta_1 = unittest.mock.Mock(
        reasoning_content="",
        content=None,
        tool_calls=None,
    )

    mock_delta_2 = unittest.mock.Mock(
        reasoning_content="\nI'm thinking",
        content=None,
        tool_calls=None,
    )
    mock_delta_3 = unittest.mock.Mock(
        reasoning_content=None,
        content="One second",
        tool_calls=None,
    )
    mock_delta_4 = unittest.mock.Mock(
        reasoning_content="\nI'm think",
        content=None,
        tool_calls=None,
    )
    mock_delta_5 = unittest.mock.Mock(
        reasoning_content="ing again",
        content=None,
        tool_calls=None,
    )

    mock_tool_call_1_part_1 = unittest.mock.Mock(index=0)
    mock_tool_call_2_part_1 = unittest.mock.Mock(index=1)
    mock_delta_6 = unittest.mock.Mock(
        content="I'll calculate", tool_calls=[mock_tool_call_1_part_1, mock_tool_call_2_part_1], reasoning_content=None
    )

    mock_tool_call_1_part_2 = unittest.mock.Mock(index=0)
    mock_tool_call_2_part_2 = unittest.mock.Mock(index=1)
    mock_delta_7 = unittest.mock.Mock(
        content="that for you", tool_calls=[mock_tool_call_1_part_2, mock_tool_call_2_part_2], reasoning_content=None
    )

    mock_delta_8 = unittest.mock.Mock(content="", tool_calls=None, reasoning_content=None)

    mock_event_1 = unittest.mock.Mock(choices=[unittest.mock.Mock(finish_reason=None, delta=mock_delta_1)])
    mock_event_2 = unittest.mock.Mock(choices=[unittest.mock.Mock(finish_reason=None, delta=mock_delta_2)])
    mock_event_3 = unittest.mock.Mock(choices=[unittest.mock.Mock(finish_reason=None, delta=mock_delta_3)])
    mock_event_4 = unittest.mock.Mock(choices=[unittest.mock.Mock(finish_reason=None, delta=mock_delta_4)])
    mock_event_5 = unittest.mock.Mock(choices=[unittest.mock.Mock(finish_reason=None, delta=mock_delta_5)])
    mock_event_6 = unittest.mock.Mock(choices=[unittest.mock.Mock(finish_reason=None, delta=mock_delta_6)])
    mock_event_7 = unittest.mock.Mock(choices=[unittest.mock.Mock(finish_reason=None, delta=mock_delta_7)])
    mock_event_8 = unittest.mock.Mock(choices=[unittest.mock.Mock(finish_reason="tool_calls", delta=mock_delta_8)])
    mock_event_9 = unittest.mock.Mock()
    mock_event_9.usage.prompt_tokens_details.cached_tokens = 10
    mock_event_9.usage.cache_creation_input_tokens = 10

    litellm_acompletion.side_effect = unittest.mock.AsyncMock(
        return_value=agenerator(
            [
                mock_event_1,
                mock_event_2,
                mock_event_3,
                mock_event_4,
                mock_event_5,
                mock_event_6,
                mock_event_7,
                mock_event_8,
                mock_event_9,
            ]
        )
    )

    messages = [{"role": "user", "content": [{"type": "text", "text": "calculate 2+2"}]}]
    response = model.stream(messages)
    tru_events = await alist(response)
    exp_events = [
        {"messageStart": {"role": "assistant"}},
        {"contentBlockStart": {"start": {}}},
        {"contentBlockDelta": {"delta": {"reasoningContent": {"text": "\nI'm thinking"}}}},
        {"contentBlockStop": {}},
        {"contentBlockStart": {"start": {}}},
        {"contentBlockDelta": {"delta": {"text": "One second"}}},
        {"contentBlockStop": {}},
        {"contentBlockStart": {"start": {}}},
        {"contentBlockDelta": {"delta": {"reasoningContent": {"text": "\nI'm think"}}}},
        {"contentBlockDelta": {"delta": {"reasoningContent": {"text": "ing again"}}}},
        {"contentBlockStop": {}},
        {"contentBlockStart": {"start": {}}},
        {"contentBlockDelta": {"delta": {"text": "I'll calculate"}}},
        {"contentBlockDelta": {"delta": {"text": "that for you"}}},
        {"contentBlockStop": {}},
        {
            "contentBlockStart": {
                "start": {
                    "toolUse": {"name": mock_tool_call_1_part_1.function.name, "toolUseId": mock_tool_call_1_part_1.id}
                }
            }
        },
        {"contentBlockDelta": {"delta": {"toolUse": {"input": mock_tool_call_1_part_1.function.arguments}}}},
        {"contentBlockDelta": {"delta": {"toolUse": {"input": mock_tool_call_1_part_2.function.arguments}}}},
        {"contentBlockStop": {}},
        {
            "contentBlockStart": {
                "start": {
                    "toolUse": {"name": mock_tool_call_2_part_1.function.name, "toolUseId": mock_tool_call_2_part_1.id}
                }
            }
        },
        {"contentBlockDelta": {"delta": {"toolUse": {"input": mock_tool_call_2_part_1.function.arguments}}}},
        {"contentBlockDelta": {"delta": {"toolUse": {"input": mock_tool_call_2_part_2.function.arguments}}}},
        {"contentBlockStop": {}},
        {"messageStop": {"stopReason": "tool_use"}},
        {
            "metadata": {
                "usage": {
                    "cacheReadInputTokens": mock_event_9.usage.prompt_tokens_details.cached_tokens,
                    "cacheWriteInputTokens": mock_event_9.usage.cache_creation_input_tokens,
                    "inputTokens": mock_event_9.usage.prompt_tokens,
                    "outputTokens": mock_event_9.usage.completion_tokens,
                    "totalTokens": mock_event_9.usage.total_tokens,
                },
                "metrics": {"latencyMs": 0},
            }
        },
    ]

    assert tru_events == exp_events

    assert litellm_acompletion.call_args_list == [
        call(
            api_key=api_key,
            messages=[{"role": "user", "content": [{"text": "calculate 2+2", "type": "text"}]}],
            model=model_id,
            stream=True,
            stream_options={"include_usage": True},
            tools=[],
        )
    ]


@pytest.mark.asyncio
async def test_stream_empty(litellm_acompletion, api_key, model_id, model, agenerator, alist):
    mock_delta = unittest.mock.Mock(content=None, tool_calls=None, reasoning_content=None)

    mock_event_1 = unittest.mock.Mock(choices=[unittest.mock.Mock(finish_reason=None, delta=mock_delta)])
    mock_event_2 = unittest.mock.Mock(choices=[unittest.mock.Mock(finish_reason="stop", delta=mock_delta)])
    mock_event_3 = unittest.mock.Mock(usage=None)
    mock_event_4 = unittest.mock.Mock(usage=None)

    litellm_acompletion.side_effect = unittest.mock.AsyncMock(
        return_value=agenerator([mock_event_1, mock_event_2, mock_event_3, mock_event_4])
    )

    messages = [{"role": "user", "content": []}]
    response = model.stream(messages)

    tru_events = await alist(response)
    exp_events = [
        {"messageStart": {"role": "assistant"}},
        {"messageStop": {"stopReason": "end_turn"}},
    ]

    assert len(tru_events) == len(exp_events)
    expected_request = {
        "api_key": api_key,
        "model": model_id,
        "messages": [],
        "stream": True,
        "stream_options": {"include_usage": True},
        "tools": [],
    }
    litellm_acompletion.assert_called_once_with(**expected_request)


@pytest.mark.asyncio
async def test_structured_output(litellm_acompletion, model, test_output_model_cls, alist):
    messages = [{"role": "user", "content": [{"text": "Generate a person"}]}]

    mock_choice = unittest.mock.Mock()
    mock_choice.finish_reason = "tool_calls"
    mock_choice.message.content = '{"name": "John", "age": 30}'
    # PATCH START: mock tool_calls as list with .function.arguments
    tool_call_mock = unittest.mock.Mock()
    tool_call_function_mock = unittest.mock.Mock()
    tool_call_function_mock.arguments = '{"name": "John", "age": 30}'
    tool_call_mock.function = tool_call_function_mock
    mock_choice.message.tool_calls = [tool_call_mock]
    # PATCH END
    mock_response = unittest.mock.Mock()
    mock_response.choices = [mock_choice]

    litellm_acompletion.side_effect = unittest.mock.AsyncMock(return_value=mock_response)

    with unittest.mock.patch.object(strands.models.litellm, "supports_response_schema", return_value=True):
        stream = model.structured_output(test_output_model_cls, messages)
        events = await alist(stream)
        tru_result = events[-1]

    exp_result = {"output": test_output_model_cls(name="John", age=30)}
    assert tru_result == exp_result


@pytest.mark.asyncio
async def test_structured_output_unsupported_model(litellm_acompletion, model, test_output_model_cls, alist):
    messages = [{"role": "user", "content": [{"text": "Generate a person"}]}]

    mock_tool_call = unittest.mock.Mock()
    mock_tool_call.function.arguments = '{"name": "John", "age": 30}'

    mock_choice = unittest.mock.Mock()
    mock_choice.finish_reason = "tool_calls"
    mock_choice.message.tool_calls = [mock_tool_call]
    mock_response = unittest.mock.Mock()
    mock_response.choices = [mock_choice]

    litellm_acompletion.return_value = mock_response

    with unittest.mock.patch.object(strands.models.litellm, "supports_response_schema", return_value=False):
        stream = model.structured_output(test_output_model_cls, messages)
        events = await alist(stream)
        tru_result = events[-1]

    exp_result = {"output": test_output_model_cls(name="John", age=30)}
    assert tru_result == exp_result


def test_config_validation_warns_on_unknown_keys(litellm_acompletion, captured_warnings):
    """Test that unknown config keys emit a warning."""
    LiteLLMModel(client_args={"api_key": "test"}, model_id="test-model", invalid_param="test")

    assert len(captured_warnings) == 1
    assert "Invalid configuration parameters" in str(captured_warnings[0].message)
    assert "invalid_param" in str(captured_warnings[0].message)


def test_update_config_validation_warns_on_unknown_keys(model, captured_warnings):
    """Test that update_config warns on unknown keys."""
    model.update_config(wrong_param="test")

    assert len(captured_warnings) == 1
    assert "Invalid configuration parameters" in str(captured_warnings[0].message)
    assert "wrong_param" in str(captured_warnings[0].message)


def test_tool_choice_supported_no_warning(model, messages, captured_warnings):
    """Test that toolChoice doesn't emit warning for supported providers."""
    tool_choice = {"auto": {}}
    model.format_request(messages, tool_choice=tool_choice)

    assert len(captured_warnings) == 0


def test_tool_choice_none_no_warning(model, messages, captured_warnings):
    """Test that None toolChoice doesn't emit warning."""
    model.format_request(messages, tool_choice=None)

    assert len(captured_warnings) == 0


@pytest.mark.asyncio
async def test_context_window_maps_to_typed_exception(litellm_acompletion, model):
    """Test that a typed ContextWindowExceededError is mapped correctly."""
    litellm_acompletion.side_effect = ContextWindowExceededError(message="test error", model="x", llm_provider="y")

    with pytest.raises(ContextWindowOverflowException):
        async for _ in model.stream([{"role": "user", "content": [{"text": "x"}]}]):
            pass


def test_format_request_messages_with_system_prompt_content():
    """Test format_request_messages with system_prompt_content parameter."""
    messages = [{"role": "user", "content": [{"text": "Hello"}]}]
    system_prompt_content = [{"text": "You are a helpful assistant."}, {"cachePoint": {"type": "default"}}]

    result = LiteLLMModel.format_request_messages(messages, system_prompt_content=system_prompt_content)

    expected = [
        {
            "role": "system",
            "content": [
                {"type": "text", "text": "You are a helpful assistant.", "cache_control": {"type": "ephemeral"}}
            ],
        },
        {"role": "user", "content": [{"text": "Hello", "type": "text"}]},
    ]

    assert result == expected


def test_format_request_messages_backward_compatibility_system_prompt():
    """Test that system_prompt is converted to system_prompt_content when system_prompt_content is None."""
    messages = [{"role": "user", "content": [{"text": "Hello"}]}]
    system_prompt = "You are a helpful assistant."

    result = LiteLLMModel.format_request_messages(messages, system_prompt=system_prompt)

    expected = [
        {"role": "system", "content": [{"type": "text", "text": "You are a helpful assistant."}]},
        {"role": "user", "content": [{"text": "Hello", "type": "text"}]},
    ]

    assert result == expected


def test_format_request_messages_cache_point_support():
    """Test that cache points are properly applied to preceding content blocks."""
    messages = [{"role": "user", "content": [{"text": "Hello"}]}]
    system_prompt_content = [
        {"text": "First instruction."},
        {"text": "Second instruction."},
        {"cachePoint": {"type": "default"}},
        {"text": "Third instruction."},
    ]

    result = LiteLLMModel.format_request_messages(messages, system_prompt_content=system_prompt_content)

    expected = [
        {
            "role": "system",
            "content": [
                {"type": "text", "text": "First instruction."},
                {"type": "text", "text": "Second instruction.", "cache_control": {"type": "ephemeral"}},
                {"type": "text", "text": "Third instruction."},
            ],
        },
        {"role": "user", "content": [{"text": "Hello", "type": "text"}]},
    ]

    assert result == expected


@pytest.mark.asyncio
async def test_stream_non_streaming(litellm_acompletion, api_key, model_id, alist):
    """Test LiteLLM model with streaming disabled (stream=False).

    This test verifies that the LiteLLM model works correctly when streaming is disabled,
    which was the issue reported in GitHub issue #477.
    """

    mock_function = unittest.mock.Mock()
    mock_function.name = "calculator"
    mock_function.arguments = '{"expression": "123981723 + 234982734"}'

    mock_tool_call = unittest.mock.Mock(index=0, function=mock_function, id="tool_call_id_123")

    mock_message = unittest.mock.Mock()
    mock_message.content = "I'll calculate that for you"
    mock_message.reasoning_content = "Let me think about this calculation"
    mock_message.tool_calls = [mock_tool_call]

    mock_choice = unittest.mock.Mock()
    mock_choice.message = mock_message
    mock_choice.finish_reason = "tool_calls"

    mock_response = unittest.mock.Mock()
    mock_response.choices = [mock_choice]

    # Create a more explicit usage mock that doesn't have cache-related attributes
    mock_usage = unittest.mock.Mock()
    mock_usage.prompt_tokens = 10
    mock_usage.completion_tokens = 20
    mock_usage.total_tokens = 30
    mock_usage.prompt_tokens_details = None
    mock_usage.cache_creation_input_tokens = None
    mock_response.usage = mock_usage

    litellm_acompletion.side_effect = unittest.mock.AsyncMock(return_value=mock_response)

    model = LiteLLMModel(
        client_args={"api_key": api_key},
        model_id=model_id,
        params={"stream": False},  # This is the key setting that was causing the #477 isuue
    )

    messages = [{"role": "user", "content": [{"type": "text", "text": "What is 123981723 + 234982734?"}]}]
    response = model.stream(messages)

    tru_events = await alist(response)

    exp_events = [
        {"messageStart": {"role": "assistant"}},
        {"contentBlockStart": {"start": {}}},
        {"contentBlockDelta": {"delta": {"reasoningContent": {"text": "Let me think about this calculation"}}}},
        {"contentBlockStop": {}},
        {"contentBlockStart": {"start": {}}},
        {"contentBlockDelta": {"delta": {"text": "I'll calculate that for you"}}},
        {"contentBlockStop": {}},
        {
            "contentBlockStart": {
                "start": {"toolUse": {"name": "calculator", "toolUseId": mock_message.tool_calls[0].id}}
            }
        },
        {"contentBlockDelta": {"delta": {"toolUse": {"input": '{"expression": "123981723 + 234982734"}'}}}},
        {"contentBlockStop": {}},
        {"messageStop": {"stopReason": "tool_use"}},
        {
            "metadata": {
                "usage": {
                    "inputTokens": 10,
                    "outputTokens": 20,
                    "totalTokens": 30,
                },
                "metrics": {"latencyMs": 0},
            }
        },
    ]

    assert len(tru_events) == len(exp_events)

    for i, (tru, exp) in enumerate(zip(tru_events, exp_events, strict=False)):
        assert tru == exp, f"Event {i} mismatch: {tru} != {exp}"

    expected_request = {
        "api_key": api_key,
        "model": model_id,
        "messages": [{"role": "user", "content": [{"text": "What is 123981723 + 234982734?", "type": "text"}]}],
        "stream": False,  # Verify that stream=False was passed to litellm
        "stream_options": {"include_usage": True},
        "tools": [],
    }
    litellm_acompletion.assert_called_once_with(**expected_request)


@pytest.mark.asyncio
async def test_stream_path_validation(litellm_acompletion, api_key, model_id, model, agenerator, alist):
    """Test that we're taking the correct streaming path and validate stream parameter."""
    mock_delta = unittest.mock.Mock(content=None, tool_calls=None, reasoning_content=None)
    mock_event_1 = unittest.mock.Mock(choices=[unittest.mock.Mock(finish_reason="stop", delta=mock_delta)])
    mock_event_2 = unittest.mock.Mock(usage=None)

    litellm_acompletion.side_effect = unittest.mock.AsyncMock(return_value=agenerator([mock_event_1, mock_event_2]))

    messages = [{"role": "user", "content": []}]
    response = model.stream(messages)

    # Consume the response
    await alist(response)

    # Validate that litellm.acompletion was called with the expected parameters
    call_args = litellm_acompletion.call_args
    assert call_args is not None, "litellm.acompletion should have been called"

    # Check if stream parameter is being set
    called_kwargs = call_args.kwargs

    # Validate we're going down the streaming path (should have stream=True)
    assert called_kwargs.get("stream") is True, f"Expected stream=True, got {called_kwargs.get('stream')}"


def test_format_request_message_content_reasoning():
    """Test formatting reasoning content."""
    content = {"reasoningContent": {"reasoningText": {"signature": "test_sig", "text": "test_thinking"}}}

    result = LiteLLMModel.format_request_message_content(content)
    expected = {"signature": "test_sig", "thinking": "test_thinking", "type": "thinking"}

    assert result == expected


def test_format_request_message_content_video():
    """Test formatting video content."""
    content = {"video": {"source": {"bytes": "base64videodata"}}}

    result = LiteLLMModel.format_request_message_content(content)
    expected = {"type": "video_url", "video_url": {"detail": "auto", "url": "base64videodata"}}

    assert result == expected


def test_apply_proxy_prefix_with_use_litellm_proxy():
    """Test _apply_proxy_prefix when use_litellm_proxy is True."""
    model = LiteLLMModel(client_args={"use_litellm_proxy": True}, model_id="openai/gpt-4")

    assert model.get_config()["model_id"] == "litellm_proxy/openai/gpt-4"


def test_apply_proxy_prefix_already_has_prefix():
    """Test _apply_proxy_prefix when model_id already has prefix."""
    model = LiteLLMModel(client_args={"use_litellm_proxy": True}, model_id="litellm_proxy/openai/gpt-4")

    # Should not add another prefix
    assert model.get_config()["model_id"] == "litellm_proxy/openai/gpt-4"


def test_apply_proxy_prefix_disabled():
    """Test _apply_proxy_prefix when use_litellm_proxy is False."""
    model = LiteLLMModel(client_args={"use_litellm_proxy": False}, model_id="openai/gpt-4")

    assert model.get_config()["model_id"] == "openai/gpt-4"


def test_format_chunk_metadata_with_cache_tokens():
    """Test format_chunk for metadata with cache tokens."""
    model = LiteLLMModel(model_id="test")

    # Mock usage data with cache tokens
    mock_usage = unittest.mock.Mock()
    mock_usage.prompt_tokens = 100
    mock_usage.completion_tokens = 50
    mock_usage.total_tokens = 150

    # Mock cache-related attributes
    mock_tokens_details = unittest.mock.Mock()
    mock_tokens_details.cached_tokens = 25
    mock_usage.prompt_tokens_details = mock_tokens_details
    mock_usage.cache_creation_input_tokens = 10

    event = {"chunk_type": "metadata", "data": mock_usage}

    result = model.format_chunk(event)

    assert result["metadata"]["usage"]["inputTokens"] == 100
    assert result["metadata"]["usage"]["outputTokens"] == 50
    assert result["metadata"]["usage"]["totalTokens"] == 150
    assert result["metadata"]["usage"]["cacheReadInputTokens"] == 25
    assert result["metadata"]["usage"]["cacheWriteInputTokens"] == 10


def test_format_chunk_metadata_without_cache_tokens():
    """Test format_chunk for metadata without cache tokens."""
    model = LiteLLMModel(model_id="test")

    # Mock usage data without cache tokens
    mock_usage = unittest.mock.Mock()
    mock_usage.prompt_tokens = 100
    mock_usage.completion_tokens = 50
    mock_usage.total_tokens = 150
    mock_usage.prompt_tokens_details = None
    mock_usage.cache_creation_input_tokens = None

    event = {"chunk_type": "metadata", "data": mock_usage}

    result = model.format_chunk(event)

    assert result["metadata"]["usage"]["inputTokens"] == 100
    assert result["metadata"]["usage"]["outputTokens"] == 50
    assert result["metadata"]["usage"]["totalTokens"] == 150
    assert "cacheReadInputTokens" not in result["metadata"]["usage"]
    assert "cacheWriteInputTokens" not in result["metadata"]["usage"]


def test_stream_switch_content_same_type():
    """Test _stream_switch_content when data_type is the same as prev_data_type."""
    model = LiteLLMModel(model_id="test")

    chunks, data_type = model._stream_switch_content("text", "text")

    assert chunks == []
    assert data_type == "text"


def test_stream_switch_content_different_type_with_prev():
    """Test _stream_switch_content when switching from one type to another."""
    model = LiteLLMModel(model_id="test")

    chunks, data_type = model._stream_switch_content("text", "reasoning_content")

    assert len(chunks) == 2
    assert chunks[0]["contentBlockStop"] == {}
    assert chunks[1]["contentBlockStart"] == {"start": {}}
    assert data_type == "text"


def test_stream_switch_content_different_type_no_prev():
    """Test _stream_switch_content when switching to a type with no previous type."""
    model = LiteLLMModel(model_id="test")

    chunks, data_type = model._stream_switch_content("text", None)

    assert len(chunks) == 1
    assert chunks[0]["contentBlockStart"] == {"start": {}}
    assert data_type == "text"


@pytest.mark.asyncio
async def test_stream_with_events_missing_usage_attribute(
    litellm_acompletion, api_key, model_id, model, agenerator, alist
):
    """Test streaming handles events that don't have a usage attribute.

    This test verifies the fix for a bug where ModelResponseStream objects
    (which don't have a 'usage' attribute) would cause an AttributeError
    when the code tried to access event.usage directly instead of using getattr.

    The bug occurred because:
    1. ModelResponse (non-streaming) has a 'usage' attribute
    2. ModelResponseStream (streaming chunks) does NOT have a 'usage' attribute
    3. The code assumed all events would have the 'usage' attribute

    Regression test for: 'ModelResponseStream' object has no attribute 'usage'
    """

    # Use spec to ensure mock objects only have specified attributes
    # This mimics the real ModelResponseStream which doesn't have 'usage'
    class MockStreamChunk:
        """Mock that mimics ModelResponseStream - no usage attribute."""

        def __init__(self, choices=None):
            self.choices = choices or []

    mock_delta = unittest.mock.Mock(content="Hello", tool_calls=None, reasoning_content=None)
    mock_event_1 = MockStreamChunk(choices=[unittest.mock.Mock(finish_reason=None, delta=mock_delta)])
    mock_event_2 = MockStreamChunk(choices=[unittest.mock.Mock(finish_reason="stop", delta=mock_delta)])
    # After finish_reason is received, remaining events in the stream also don't have 'usage'
    mock_event_3 = MockStreamChunk(choices=[])
    mock_event_4 = MockStreamChunk(choices=[])

    litellm_acompletion.side_effect = unittest.mock.AsyncMock(
        return_value=agenerator([mock_event_1, mock_event_2, mock_event_3, mock_event_4])
    )

    messages = [{"role": "user", "content": [{"type": "text", "text": "Hello"}]}]
    response = model.stream(messages)

    # This should NOT raise AttributeError: 'MockStreamChunk' object has no attribute 'usage'
    tru_events = await alist(response)

    # Verify we got the expected events (no metadata since no usage was available)
    assert tru_events[0] == {"messageStart": {"role": "assistant"}}
    assert {"messageStop": {"stopReason": "end_turn"}} in tru_events
    # No metadata event since mock events don't have usage
    assert not any("metadata" in event for event in tru_events)


@pytest.mark.asyncio
async def test_stream_with_usage_in_final_event(litellm_acompletion, api_key, model_id, model, agenerator, alist):
    """Test streaming correctly extracts usage when it IS present in final events.

    This test ensures that when usage data IS available (e.g., with stream_options.include_usage=True),
    it is correctly extracted and included in the metadata event.
    """

    class MockStreamChunkWithoutUsage:
        """Mock streaming chunk without usage."""

        def __init__(self, choices=None):
            self.choices = choices or []

    class MockStreamChunkWithUsage:
        """Mock streaming chunk with usage (final event)."""

        def __init__(self, usage):
            self.choices = []
            self.usage = usage

    mock_delta = unittest.mock.Mock(content="Hi", tool_calls=None, reasoning_content=None)
    mock_event_1 = MockStreamChunkWithoutUsage(choices=[unittest.mock.Mock(finish_reason=None, delta=mock_delta)])
    mock_event_2 = MockStreamChunkWithoutUsage(choices=[unittest.mock.Mock(finish_reason="stop", delta=mock_delta)])

    # Final event with usage data
    mock_usage = unittest.mock.Mock()
    mock_usage.prompt_tokens = 10
    mock_usage.completion_tokens = 5
    mock_usage.total_tokens = 15
    mock_usage.prompt_tokens_details = None
    mock_usage.cache_creation_input_tokens = None
    mock_event_3 = MockStreamChunkWithUsage(usage=mock_usage)

    litellm_acompletion.side_effect = unittest.mock.AsyncMock(
        return_value=agenerator([mock_event_1, mock_event_2, mock_event_3])
    )

    messages = [{"role": "user", "content": [{"type": "text", "text": "Hi"}]}]
    response = model.stream(messages)

    tru_events = await alist(response)

    # Verify metadata event is present with correct usage
    metadata_events = [e for e in tru_events if "metadata" in e]
    assert len(metadata_events) == 1
    assert metadata_events[0]["metadata"]["usage"]["inputTokens"] == 10
    assert metadata_events[0]["metadata"]["usage"]["outputTokens"] == 5
    assert metadata_events[0]["metadata"]["usage"]["totalTokens"] == 15


def test_format_request_messages_with_tool_calls_no_content():
    """Test that assistant messages with only tool calls are included and have no content field."""
    messages = [
        {"role": "user", "content": [{"text": "Use the calculator"}]},
        {
            "role": "assistant",
            "content": [
                {
                    "toolUse": {
                        "input": {"expression": "2+2"},
                        "name": "calculator",
                        "toolUseId": "c1",
                    },
                },
            ],
        },
    ]

    tru_result = LiteLLMModel.format_request_messages(messages)

    exp_result = [
        {"role": "user", "content": [{"text": "Use the calculator", "type": "text"}]},
        {
            "role": "assistant",
            "tool_calls": [
                {
                    "function": {"arguments": '{"expression": "2+2"}', "name": "calculator"},
                    "id": "c1",
                    "type": "function",
                }
            ],
        },
    ]
    assert tru_result == exp_result



#!/usr/bin/env python3
"""Integration tests for tool retry hook mechanism.

Tests that setting AfterToolCallEvent.retry=True causes tool re-execution.
Uses direct tool invocation to test the executor-level retry, not model behavior.
"""

from strands import Agent, tool
from strands.hooks import AfterToolCallEvent


def test_tool_retry_hook_causes_reexecution():
    """Test that setting retry=True on AfterToolCallEvent causes tool re-execution.

    Verifies:
    1. Tool is called again when retry=True
    2. Hook receives AfterToolCallEvent for BOTH attempts
    3. Same tool_use_id is used (proves executor retry, not model re-calling)
    """
    state = {"call_count": 0}

    @tool(name="flaky_tool")
    def flaky_tool(message: str) -> str:
        """A tool that fails once then succeeds.

        Args:
            message: A message to include in the response.
        """
        state["call_count"] += 1
        if state["call_count"] == 1:
            raise RuntimeError("First call fails")
        return f"Success on attempt {state['call_count']}"

    hook_calls: list[dict] = []

    def retry_on_first_error(event: AfterToolCallEvent) -> None:
        tool_use_id = str(event.tool_use.get("toolUseId", ""))
        hook_calls.append(
            {
                "tool_use_id": tool_use_id,
                "status": event.result.get("status"),
                "attempt": state["call_count"],
            }
        )

        # Retry once on error
        if event.result.get("status") == "error" and state["call_count"] == 1:
            event.retry = True

    agent = Agent(tools=[flaky_tool])
    agent.hooks.add_callback(AfterToolCallEvent, retry_on_first_error)

    # Direct tool invocation bypasses model - tests executor retry mechanism
    result = agent.tool.flaky_tool(message="test")

    # Tool was called twice (1 failure + 1 success)
    assert state["call_count"] == 2

    # Hook received AfterToolCallEvent for BOTH attempts
    assert len(hook_calls) == 2
    assert hook_calls[0]["status"] == "error"
    assert hook_calls[0]["attempt"] == 1
    assert hook_calls[1]["status"] == "success"
    assert hook_calls[1]["attempt"] == 2

    # Both calls used the same tool_use_id (executor retry, not new model call)
    assert hook_calls[0]["tool_use_id"] == hook_calls[1]["tool_use_id"]

    assert result["status"] == "success"



"""Integration tests for MCP task-augmented tool execution."""

import os
import socket
import threading
import time
from typing import Any

import pytest
from mcp.client.streamable_http import streamablehttp_client

from strands.tools.mcp import MCPClient, MCPTransport, TasksConfig


def _find_available_port() -> int:
    """Find an available port."""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("127.0.0.1", 0))
        s.listen(1)
        return s.getsockname()[1]


def start_task_server(port: int) -> None:
    """Start the task echo server in a thread."""
    import uvicorn

    from tests_integ.mcp.task_echo_server import create_starlette_app

    starlette_app, _ = create_starlette_app(port)
    uvicorn.run(starlette_app, host="127.0.0.1", port=port, log_level="warning")


@pytest.fixture(scope="module")
def task_server_port() -> int:
    return _find_available_port()


@pytest.fixture(scope="module")
def task_server(task_server_port: int) -> Any:
    """Start the task server for the test module."""
    server_thread = threading.Thread(target=start_task_server, kwargs={"port": task_server_port}, daemon=True)
    server_thread.start()
    time.sleep(2)
    yield


@pytest.fixture
def task_mcp_client(task_server: Any, task_server_port: int) -> MCPClient:
    """Create an MCP client with tasks enabled."""

    def transport_callback() -> MCPTransport:
        return streamablehttp_client(url=f"http://127.0.0.1:{task_server_port}/mcp")

    return MCPClient(transport_callback, tasks_config=TasksConfig())


@pytest.fixture
def task_mcp_client_disabled(task_server: Any, task_server_port: int) -> MCPClient:
    """Create an MCP client with tasks disabled (default)."""

    def transport_callback() -> MCPTransport:
        return streamablehttp_client(url=f"http://127.0.0.1:{task_server_port}/mcp")

    return MCPClient(transport_callback)


@pytest.mark.skipif(os.environ.get("GITHUB_ACTIONS") == "true", reason="streamable transport failing in CI")
class TestMCPTaskSupport:
    """Integration tests for MCP task-augmented execution."""

    def test_direct_call_tools(self, task_mcp_client: MCPClient) -> None:
        """Test tools that use direct call_tool (forbidden or no taskSupport)."""
        with task_mcp_client:
            task_mcp_client.list_tools_sync()

            # Tool with taskSupport='forbidden'
            r1 = task_mcp_client.call_tool_sync(
                tool_use_id="t1", name="task_forbidden_echo", arguments={"message": "Hello!"}
            )
            assert r1["status"] == "success"
            assert "Forbidden echo: Hello!" in r1["content"][0].get("text", "")

            # Tool without taskSupport
            r2 = task_mcp_client.call_tool_sync(tool_use_id="t2", name="echo", arguments={"message": "Simple!"})
            assert r2["status"] == "success"
            assert "Simple echo: Simple!" in r2["content"][0].get("text", "")

    def test_task_augmented_tools(self, task_mcp_client: MCPClient) -> None:
        """Test tools that use task-augmented execution (required or optional)."""
        with task_mcp_client:
            task_mcp_client.list_tools_sync()

            # Tool with taskSupport='required'
            r1 = task_mcp_client.call_tool_sync(
                tool_use_id="t1", name="task_required_echo", arguments={"message": "Required!"}
            )
            assert r1["status"] == "success"
            assert "Task echo: Required!" in r1["content"][0].get("text", "")

            # Tool with taskSupport='optional'
            r2 = task_mcp_client.call_tool_sync(
                tool_use_id="t2", name="task_optional_echo", arguments={"message": "Optional!"}
            )
            assert r2["status"] == "success"
            assert "Task optional echo: Optional!" in r2["content"][0].get("text", "")

    def test_task_support_tool_detection(self, task_mcp_client: MCPClient) -> None:
        """Test tool-level task support detection."""
        with task_mcp_client:
            task_mcp_client.list_tools_sync()

            # Verify decision logic
            assert task_mcp_client._should_use_task("task_required_echo") is True
            assert task_mcp_client._should_use_task("task_optional_echo") is True
            assert task_mcp_client._should_use_task("task_forbidden_echo") is False
            assert task_mcp_client._should_use_task("echo") is False

    def test_server_capabilities(self, task_mcp_client: MCPClient) -> None:
        """Test server task capability detection."""
        with task_mcp_client:
            task_mcp_client.list_tools_sync()
            assert task_mcp_client._has_server_task_support() is True

    def test_tasks_disabled_by_default(self, task_mcp_client_disabled: MCPClient) -> None:
        """Test that tasks are disabled when experimental.tasks is not configured."""
        with task_mcp_client_disabled:
            task_mcp_client_disabled.list_tools_sync()

            assert task_mcp_client_disabled._is_tasks_enabled() is False
            assert task_mcp_client_disabled._should_use_task("task_required_echo") is False

            # Direct call_tool still works for tools that support it
            result = task_mcp_client_disabled.call_tool_sync(
                tool_use_id="t", name="task_optional_echo", arguments={"message": "Direct!"}
            )
            assert result["status"] == "success"

            # Task-required tools fail gracefully via direct call
            result2 = task_mcp_client_disabled.call_tool_sync(
                tool_use_id="t2", name="task_required_echo", arguments={"message": "Direct!"}
            )
            assert result2["status"] == "error"

    @pytest.mark.asyncio
    async def test_async_tool_call(self, task_mcp_client: MCPClient) -> None:
        """Test async tool calls."""
        with task_mcp_client:
            task_mcp_client.list_tools_sync()
            result = await task_mcp_client.call_tool_async(
                tool_use_id="t", name="task_forbidden_echo", arguments={"message": "Async!"}
            )
            assert result["status"] == "success"
            assert "Forbidden echo: Async!" in result["content"][0].get("text", "")



"""Tests for A2AAgent class."""

from contextlib import asynccontextmanager
from unittest.mock import AsyncMock, MagicMock, patch
from uuid import uuid4

import pytest
from a2a.types import AgentCard, Message, Part, Role, TextPart

from strands.agent.a2a_agent import A2AAgent
from strands.agent.agent_result import AgentResult


@pytest.fixture
def mock_agent_card():
    """Mock AgentCard for testing."""
    return AgentCard(
        name="test-agent",
        description="Test agent",
        url="http://localhost:8000",
        version="1.0.0",
        capabilities={},
        default_input_modes=["text/plain"],
        default_output_modes=["text/plain"],
        skills=[],
    )


@pytest.fixture
def a2a_agent():
    """Create A2AAgent instance for testing."""
    return A2AAgent(endpoint="http://localhost:8000")


@pytest.fixture
def mock_httpx_client():
    """Create a mock httpx.AsyncClient that works as async context manager."""
    mock_client = AsyncMock()
    mock_client.__aenter__.return_value = mock_client
    mock_client.__aexit__.return_value = None
    return mock_client


@asynccontextmanager
async def mock_a2a_client_context(send_message_func):
    """Helper to create mock A2A client setup for _send_message tests."""
    mock_client = MagicMock()
    mock_client.send_message = send_message_func
    with patch("strands.agent.a2a_agent.httpx.AsyncClient") as mock_httpx_class:
        mock_httpx = AsyncMock()
        mock_httpx.__aenter__.return_value = mock_httpx
        mock_httpx.__aexit__.return_value = None
        mock_httpx_class.return_value = mock_httpx
        with patch("strands.agent.a2a_agent.ClientFactory") as mock_factory_class:
            mock_factory = MagicMock()
            mock_factory.create.return_value = mock_client
            mock_factory_class.return_value = mock_factory
            yield mock_httpx_class, mock_factory_class


def test_init_with_defaults():
    """Test initialization with default parameters."""
    agent = A2AAgent(endpoint="http://localhost:8000")
    assert agent.endpoint == "http://localhost:8000"
    assert agent.timeout == 300
    assert agent._agent_card is None
    assert agent.name is None
    assert agent.description is None


def test_init_with_name_and_description():
    """Test initialization with custom name and description."""
    agent = A2AAgent(endpoint="http://localhost:8000", name="my-agent", description="My custom agent")
    assert agent.name == "my-agent"
    assert agent.description == "My custom agent"


def test_init_with_custom_timeout():
    """Test initialization with custom timeout."""
    agent = A2AAgent(endpoint="http://localhost:8000", timeout=600)
    assert agent.timeout == 600


def test_init_with_external_a2a_client_factory():
    """Test initialization with external A2A client factory."""
    external_factory = MagicMock()
    agent = A2AAgent(endpoint="http://localhost:8000", a2a_client_factory=external_factory)
    assert agent._a2a_client_factory is external_factory


@pytest.mark.asyncio
async def test_get_agent_card(a2a_agent, mock_agent_card, mock_httpx_client):
    """Test agent card discovery."""
    with patch("strands.agent.a2a_agent.httpx.AsyncClient", return_value=mock_httpx_client):
        with patch("strands.agent.a2a_agent.A2ACardResolver") as mock_resolver_class:
            mock_resolver = AsyncMock()
            mock_resolver.get_agent_card = AsyncMock(return_value=mock_agent_card)
            mock_resolver_class.return_value = mock_resolver

            card = await a2a_agent.get_agent_card()

            assert card == mock_agent_card
            assert a2a_agent._agent_card == mock_agent_card


@pytest.mark.asyncio
async def test_get_agent_card_cached(a2a_agent, mock_agent_card):
    """Test that agent card is cached after first discovery."""
    a2a_agent._agent_card = mock_agent_card

    card = await a2a_agent.get_agent_card()

    assert card == mock_agent_card


@pytest.mark.asyncio
async def test_get_agent_card_populates_name_and_description(mock_agent_card, mock_httpx_client):
    """Test that agent card populates name and description if not set."""
    agent = A2AAgent(endpoint="http://localhost:8000")

    with patch("strands.agent.a2a_agent.httpx.AsyncClient", return_value=mock_httpx_client):
        with patch("strands.agent.a2a_agent.A2ACardResolver") as mock_resolver_class:
            mock_resolver = AsyncMock()
            mock_resolver.get_agent_card = AsyncMock(return_value=mock_agent_card)
            mock_resolver_class.return_value = mock_resolver

            await agent.get_agent_card()

            assert agent.name == mock_agent_card.name
            assert agent.description == mock_agent_card.description


@pytest.mark.asyncio
async def test_get_agent_card_preserves_custom_name_and_description(mock_agent_card, mock_httpx_client):
    """Test that custom name and description are not overridden by agent card."""
    agent = A2AAgent(endpoint="http://localhost:8000", name="custom-name", description="Custom description")

    with patch("strands.agent.a2a_agent.httpx.AsyncClient", return_value=mock_httpx_client):
        with patch("strands.agent.a2a_agent.A2ACardResolver") as mock_resolver_class:
            mock_resolver = AsyncMock()
            mock_resolver.get_agent_card = AsyncMock(return_value=mock_agent_card)
            mock_resolver_class.return_value = mock_resolver

            await agent.get_agent_card()

            assert agent.name == "custom-name"
            assert agent.description == "Custom description"


@pytest.mark.asyncio
async def test_invoke_async_success(a2a_agent, mock_agent_card):
    """Test successful async invocation."""
    mock_response = Message(
        message_id=uuid4().hex,
        role=Role.agent,
        parts=[Part(TextPart(kind="text", text="Response"))],
    )

    async def mock_send_message(*args, **kwargs):
        yield mock_response

    with patch.object(a2a_agent, "get_agent_card", return_value=mock_agent_card):
        async with mock_a2a_client_context(mock_send_message):
            result = await a2a_agent.invoke_async("Hello")

            assert isinstance(result, AgentResult)
            assert result.message["content"][0]["text"] == "Response"


@pytest.mark.asyncio
async def test_invoke_async_no_prompt(a2a_agent):
    """Test that invoke_async raises ValueError when prompt is None."""
    with pytest.raises(ValueError, match="prompt is required"):
        await a2a_agent.invoke_async(None)


@pytest.mark.asyncio
async def test_invoke_async_no_response(a2a_agent, mock_agent_card):
    """Test that invoke_async raises RuntimeError when no response received."""

    async def mock_send_message(*args, **kwargs):
        return
        yield  # Make it an async generator

    with patch.object(a2a_agent, "get_agent_card", return_value=mock_agent_card):
        async with mock_a2a_client_context(mock_send_message):
            with pytest.raises(RuntimeError, match="No response received"):
                await a2a_agent.invoke_async("Hello")


def test_call_sync(a2a_agent):
    """Test synchronous call method."""
    mock_result = AgentResult(
        stop_reason="end_turn",
        message={"role": "assistant", "content": [{"text": "Response"}]},
        metrics=MagicMock(),
        state={},
    )

    with patch("strands.agent.a2a_agent.run_async") as mock_run_async:
        mock_run_async.return_value = mock_result

        result = a2a_agent("Hello")

        assert result == mock_result
        mock_run_async.assert_called_once()


@pytest.mark.asyncio
async def test_stream_async_success(a2a_agent, mock_agent_card):
    """Test successful async streaming."""
    mock_response = Message(
        message_id=uuid4().hex,
        role=Role.agent,
        parts=[Part(TextPart(kind="text", text="Response"))],
    )

    async def mock_send_message(*args, **kwargs):
        yield mock_response

    with patch.object(a2a_agent, "get_agent_card", return_value=mock_agent_card):
        async with mock_a2a_client_context(mock_send_message):
            events = []
            async for event in a2a_agent.stream_async("Hello"):
                events.append(event)

            assert len(events) == 2
            # First event is A2A stream event
            assert events[0]["type"] == "a2a_stream"
            assert events[0]["event"] == mock_response
            # Final event is AgentResult
            assert "result" in events[1]
            assert isinstance(events[1]["result"], AgentResult)
            assert events[1]["result"].message["content"][0]["text"] == "Response"


@pytest.mark.asyncio
async def test_stream_async_no_prompt(a2a_agent):
    """Test that stream_async raises ValueError when prompt is None."""
    with pytest.raises(ValueError, match="prompt is required"):
        async for _ in a2a_agent.stream_async(None):
            pass


@pytest.mark.asyncio
async def test_send_message_uses_provided_factory(mock_agent_card):
    """Test _send_message uses provided factory instead of creating per-call client."""
    external_factory = MagicMock()
    mock_a2a_client = MagicMock()

    async def mock_send_message(*args, **kwargs):
        yield MagicMock()

    mock_a2a_client.send_message = mock_send_message
    external_factory.create.return_value = mock_a2a_client

    agent = A2AAgent(endpoint="http://localhost:8000", a2a_client_factory=external_factory)

    with patch.object(agent, "get_agent_card", return_value=mock_agent_card):
        # Consume the async iterator
        async for _ in agent._send_message("Hello"):
            pass

        external_factory.create.assert_called_once_with(mock_agent_card)


@pytest.mark.asyncio
async def test_send_message_creates_per_call_client(a2a_agent, mock_agent_card):
    """Test _send_message creates a fresh httpx client for each call when no factory provided."""
    mock_response = Message(
        message_id=uuid4().hex,
        role=Role.agent,
        parts=[Part(TextPart(kind="text", text="Response"))],
    )

    async def mock_send_message(*args, **kwargs):
        yield mock_response

    with patch.object(a2a_agent, "get_agent_card", return_value=mock_agent_card):
        async with mock_a2a_client_context(mock_send_message) as (mock_httpx_class, _):
            # Consume the async iterator
            async for _ in a2a_agent._send_message("Hello"):
                pass

            # Verify httpx client was created with timeout
            mock_httpx_class.assert_called_once_with(timeout=300)


def test_is_complete_event_message(a2a_agent):
    """Test _is_complete_event returns True for Message."""
    mock_message = MagicMock(spec=Message)

    assert a2a_agent._is_complete_event(mock_message) is True


def test_is_complete_event_tuple_with_none_update(a2a_agent):
    """Test _is_complete_event returns True for tuple with None update event."""
    mock_task = MagicMock()

    assert a2a_agent._is_complete_event((mock_task, None)) is True


def test_is_complete_event_artifact_last_chunk(a2a_agent):
    """Test _is_complete_event handles TaskArtifactUpdateEvent last_chunk flag."""
    from a2a.types import TaskArtifactUpdateEvent

    mock_task = MagicMock()

    # last_chunk=True -> complete
    event_complete = MagicMock(spec=TaskArtifactUpdateEvent)
    event_complete.last_chunk = True
    assert a2a_agent._is_complete_event((mock_task, event_complete)) is True

    # last_chunk=False -> not complete
    event_incomplete = MagicMock(spec=TaskArtifactUpdateEvent)
    event_incomplete.last_chunk = False
    assert a2a_agent._is_complete_event((mock_task, event_incomplete)) is False

    # last_chunk=None -> not complete
    event_none = MagicMock(spec=TaskArtifactUpdateEvent)
    event_none.last_chunk = None
    assert a2a_agent._is_complete_event((mock_task, event_none)) is False


def test_is_complete_event_status_update(a2a_agent):
    """Test _is_complete_event handles TaskStatusUpdateEvent state."""
    from a2a.types import TaskState, TaskStatusUpdateEvent

    mock_task = MagicMock()

    # completed state -> complete
    event_completed = MagicMock(spec=TaskStatusUpdateEvent)
    event_completed.status = MagicMock()
    event_completed.status.state = TaskState.completed
    assert a2a_agent._is_complete_event((mock_task, event_completed)) is True

    # working state -> not complete
    event_working = MagicMock(spec=TaskStatusUpdateEvent)
    event_working.status = MagicMock()
    event_working.status.state = TaskState.working
    assert a2a_agent._is_complete_event((mock_task, event_working)) is False

    # no status -> not complete
    event_no_status = MagicMock(spec=TaskStatusUpdateEvent)
    event_no_status.status = None
    assert a2a_agent._is_complete_event((mock_task, event_no_status)) is False


def test_is_complete_event_unknown_type(a2a_agent):
    """Test _is_complete_event returns False for unknown event types."""
    assert a2a_agent._is_complete_event("unknown") is False


@pytest.mark.asyncio
async def test_stream_async_tracks_complete_events(a2a_agent, mock_agent_card):
    """Test stream_async uses last complete event for final result."""
    from a2a.types import TaskState, TaskStatusUpdateEvent

    mock_task = MagicMock()
    mock_task.artifacts = None

    # First event: incomplete
    incomplete_event = MagicMock(spec=TaskStatusUpdateEvent)
    incomplete_event.status = MagicMock()
    incomplete_event.status.state = TaskState.working
    incomplete_event.status.message = None

    # Second event: complete
    complete_event = MagicMock(spec=TaskStatusUpdateEvent)
    complete_event.status = MagicMock()
    complete_event.status.state = TaskState.completed
    complete_event.status.message = MagicMock()
    complete_event.status.message.parts = []

    async def mock_send_message(*args, **kwargs):
        yield (mock_task, incomplete_event)
        yield (mock_task, complete_event)

    with patch.object(a2a_agent, "get_agent_card", return_value=mock_agent_card):
        async with mock_a2a_client_context(mock_send_message):
            events = []
            async for event in a2a_agent.stream_async("Hello"):
                events.append(event)

            # Should have 2 stream events + 1 result event
            assert len(events) == 3
            assert "result" in events[2]


@pytest.mark.asyncio
async def test_stream_async_falls_back_to_last_event(a2a_agent, mock_agent_card):
    """Test stream_async falls back to last event when no complete event."""
    from a2a.types import TaskState, TaskStatusUpdateEvent

    mock_task = MagicMock()
    mock_task.artifacts = None

    incomplete_event = MagicMock(spec=TaskStatusUpdateEvent)
    incomplete_event.status = MagicMock()
    incomplete_event.status.state = TaskState.working
    incomplete_event.status.message = None

    async def mock_send_message(*args, **kwargs):
        yield (mock_task, incomplete_event)

    with patch.object(a2a_agent, "get_agent_card", return_value=mock_agent_card):
        async with mock_a2a_client_context(mock_send_message):
            events = []
            async for event in a2a_agent.stream_async("Hello"):
                events.append(event)

            # Should have 1 stream event + 1 result event (falls back to last)
            assert len(events) == 2
            assert "result" in events[1]



"""
Integration tests for MCP client resource functionality.

This module tests the resource-related methods in MCPClient:
- list_resources_sync()
- read_resource_sync()
- list_resource_templates_sync()

The tests use the echo server which has been extended with resource functionality.
"""

import base64
import json

import pytest
from mcp import StdioServerParameters, stdio_client
from mcp.shared.exceptions import McpError
from mcp.types import BlobResourceContents, TextResourceContents
from pydantic import AnyUrl

from strands.tools.mcp.mcp_client import MCPClient


def test_mcp_resources_list_and_read():
    """Test listing and reading various types of resources."""
    mcp_client = MCPClient(
        lambda: stdio_client(StdioServerParameters(command="python", args=["tests_integ/mcp/echo_server.py"]))
    )

    with mcp_client:
        # Test list_resources_sync
        resources_result = mcp_client.list_resources_sync()
        assert len(resources_result.resources) >= 2  # At least our 2 static resources

        # Verify resource URIs exist (only static resources, not templates)
        resource_uris = [str(r.uri) for r in resources_result.resources]
        assert "test://static-text" in resource_uris
        assert "test://static-binary" in resource_uris
        # Template resources are not listed in static resources

        # Test reading text resource
        text_resource = mcp_client.read_resource_sync("test://static-text")
        assert len(text_resource.contents) == 1
        content = text_resource.contents[0]
        assert isinstance(content, TextResourceContents)
        assert "This is the content of the static text resource." in content.text

        # Test reading binary resource
        binary_resource = mcp_client.read_resource_sync("test://static-binary")
        assert len(binary_resource.contents) == 1
        binary_content = binary_resource.contents[0]
        assert isinstance(binary_content, BlobResourceContents)
        # Verify it's valid base64 encoded data
        decoded_data = base64.b64decode(binary_content.blob)
        assert len(decoded_data) > 0


def test_mcp_resources_templates():
    """Test listing resource templates and reading from template resources."""
    mcp_client = MCPClient(
        lambda: stdio_client(StdioServerParameters(command="python", args=["tests_integ/mcp/echo_server.py"]))
    )

    with mcp_client:
        # Test list_resource_templates_sync
        templates_result = mcp_client.list_resource_templates_sync()
        assert len(templates_result.resourceTemplates) >= 1

        # Verify template URIs exist
        template_uris = [t.uriTemplate for t in templates_result.resourceTemplates]
        assert "test://template/{id}/data" in template_uris

        # Test reading from template resource
        template_resource = mcp_client.read_resource_sync("test://template/123/data")
        assert len(template_resource.contents) == 1
        template_content = template_resource.contents[0]
        assert isinstance(template_content, TextResourceContents)

        # Parse the JSON response
        parsed_json = json.loads(template_content.text)
        assert parsed_json["id"] == "123"
        assert parsed_json["templateTest"] is True
        assert "Data for ID: 123" in parsed_json["data"]


def test_mcp_resources_pagination():
    """Test pagination support for resources."""
    mcp_client = MCPClient(
        lambda: stdio_client(StdioServerParameters(command="python", args=["tests_integ/mcp/echo_server.py"]))
    )

    with mcp_client:
        # Test with pagination token (should work even if server doesn't implement pagination)
        resources_result = mcp_client.list_resources_sync(pagination_token=None)
        assert len(resources_result.resources) >= 0

        # Test resource templates pagination
        templates_result = mcp_client.list_resource_templates_sync(pagination_token=None)
        assert len(templates_result.resourceTemplates) >= 0


def test_mcp_resources_error_handling():
    """Test error handling for resource operations."""
    mcp_client = MCPClient(
        lambda: stdio_client(StdioServerParameters(command="python", args=["tests_integ/mcp/echo_server.py"]))
    )

    with mcp_client:
        # Test reading non-existent resource
        with pytest.raises(McpError, match="Unknown resource"):
            mcp_client.read_resource_sync("test://nonexistent")


def test_mcp_resources_uri_types():
    """Test that both string and AnyUrl types work for read_resource_sync."""
    mcp_client = MCPClient(
        lambda: stdio_client(StdioServerParameters(command="python", args=["tests_integ/mcp/echo_server.py"]))
    )

    with mcp_client:
        # Test with string URI
        text_resource_str = mcp_client.read_resource_sync("test://static-text")
        assert len(text_resource_str.contents) == 1

        # Test with AnyUrl URI
        text_resource_url = mcp_client.read_resource_sync(AnyUrl("test://static-text"))
        assert len(text_resource_url.contents) == 1

        # Both should return the same content
        assert text_resource_str.contents[0].text == text_resource_url.contents[0].text



"""Pytest fixtures for bidirectional streaming integration tests."""

import logging

import pytest

from .generators.audio import AudioGenerator

logger = logging.getLogger(__name__)


@pytest.fixture(scope="session")
def audio_generator():
    """Provide AudioGenerator instance for tests."""
    return AudioGenerator(region="us-east-1")


@pytest.fixture(autouse=True)
def setup_logging():
    """Configure logging for tests."""
    logging.basicConfig(
        level=logging.DEBUG,
        format="%(levelname)s | %(name)s | %(message)s",
    )
    # Reduce noise from some loggers
    logging.getLogger("boto3").setLevel(logging.WARNING)
    logging.getLogger("botocore").setLevel(logging.WARNING)
    logging.getLogger("urllib3").setLevel(logging.WARNING)



"""Model-related type definitions for bidirectional streaming.

Defines types and configurations that are central to model providers,
including audio configuration that models use to specify their audio
processing requirements.
"""

from typing import TypedDict

from .events import AudioChannel, AudioFormat, AudioSampleRate


class AudioConfig(TypedDict, total=False):
    """Audio configuration for bidirectional streaming models.

    Defines standard audio parameters that model providers use to specify
    their audio processing requirements. All fields are optional to support
    models that may not use audio or only need specific parameters.

    Model providers build this configuration by merging user-provided values
    with their own defaults. The resulting configuration is then used by
    audio I/O implementations to configure hardware appropriately.

    Attributes:
        input_rate: Input sample rate in Hz (e.g., 16000, 24000, 48000)
        output_rate: Output sample rate in Hz (e.g., 16000, 24000, 48000)
        channels: Number of audio channels (1=mono, 2=stereo)
        format: Audio encoding format
        voice: Voice identifier for text-to-speech (e.g., "alloy", "matthew")
    """

    input_rate: AudioSampleRate
    output_rate: AudioSampleRate
    channels: AudioChannel
    format: AudioFormat
    voice: str



"""
Echo Server for MCP Integration Testing

This module implements a simple echo server using the Model Context Protocol (MCP).
It provides basic tools that echo back input strings and structured content, which is useful for
testing the MCP communication flow and validating that messages are properly
transmitted between the client and server.

The server runs with stdio transport, making it suitable for integration tests
where the client can spawn this process and communicate with it through standard
input/output streams.

Usage:
    Run this file directly to start the echo server:
    $ python echo_server.py
"""

import base64
import json
from typing import Literal

from mcp.server import FastMCP
from mcp.types import BlobResourceContents, CallToolResult, EmbeddedResource, TextContent, TextResourceContents
from pydantic import BaseModel

TEST_IMAGE_BASE64 = "iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAYAAAAfFcSJAAAADUlEQVR42mP8z8DwHwAFBQIAX8jx0gAAAABJRU5ErkJggg=="


class EchoResponse(BaseModel):
    """Response model for echo with structured content."""

    echoed: str
    message_length: int


def start_echo_server():
    """
    Initialize and start the MCP echo server.

    Creates a FastMCP server instance with tools that return
    input strings and structured content back to the caller. The server uses stdio transport
    for communication.

    """
    mcp = FastMCP("Echo Server")

    @mcp.tool(description="Echos response back to the user", structured_output=False)
    def echo(to_echo: str) -> str:
        return to_echo

    # FastMCP automatically constructs structured output schema from method signature
    @mcp.tool(description="Echos response back with structured content", structured_output=True)
    def echo_with_structured_content(to_echo: str) -> EchoResponse:
        return EchoResponse(echoed=to_echo, message_length=len(to_echo))

    @mcp.tool(description="Echos response back with metadata")
    def echo_with_metadata(to_echo: str):
        """Return structured content and metadata in the tool result."""

        return CallToolResult(
            content=[TextContent(type="text", text=to_echo)],
            isError=False,
            _meta={"metadata": {"nested": 1}, "shallow": "val"},
        )

    @mcp.tool(description="Get current weather information for a location")
    def get_weather(location: Literal["New York", "London", "Tokyo"] = "New York"):
        """Get weather data including forecasts and alerts for the specified location"""
        if location.lower() == "new york":
            return [
                EmbeddedResource(
                    type="resource",
                    resource=TextResourceContents(
                        uri="https://weather.api/forecast/nyc",
                        mimeType="text/plain",
                        text="Current weather in New York: 72°F, partly cloudy with light winds.",
                    ),
                )
            ]
        elif location.lower() == "london":
            return [
                EmbeddedResource(
                    type="resource",
                    resource=BlobResourceContents(
                        uri="https://weather.api/data/london.json",
                        mimeType="application/json",
                        blob=base64.b64encode(b'{"temperature": 18, "condition": "rainy", "humidity": 85}').decode(),
                    ),
                )
            ]
        elif location.lower() == "tokyo":
            # Read yellow.png file for weather icon
            with open("tests_integ/resources/yellow.png", "rb") as image_file:
                png_data = image_file.read()
            return [
                EmbeddedResource(
                    type="resource",
                    resource=BlobResourceContents(
                        uri="https://weather.api/icons/sunny.png",
                        mimeType="image/png",
                        blob=base64.b64encode(png_data).decode(),
                    ),
                )
            ]

    # Resources
    @mcp.resource("test://static-text")
    def static_text_resource() -> str:
        """A static text resource for testing"""
        return "This is the content of the static text resource."

    @mcp.resource("test://static-binary")
    def static_binary_resource() -> bytes:
        """A static binary resource (image) for testing"""
        return base64.b64decode(TEST_IMAGE_BASE64)

    @mcp.resource("test://template/{id}/data")
    def template_resource(id: str) -> str:
        """A resource template with parameter substitution"""
        return json.dumps({"id": id, "templateTest": True, "data": f"Data for ID: {id}"})

    mcp.run(transport="stdio")


if __name__ == "__main__":
    start_echo_server()



"""Tests for thinking mode behavior in BedrockModel."""

import pytest

from strands.models.bedrock import BedrockModel


@pytest.fixture
def model_with_thinking():
    """Create a BedrockModel with thinking enabled."""
    return BedrockModel(
        model_id="anthropic.claude-sonnet-4-20250514-v1:0",
        additional_request_fields={"thinking": {"type": "enabled", "budget_tokens": 5000}},
    )


@pytest.fixture
def model_without_thinking():
    """Create a BedrockModel without thinking."""
    return BedrockModel(model_id="anthropic.claude-sonnet-4-20250514-v1:0")


@pytest.fixture
def model_with_thinking_and_other_fields():
    """Create a BedrockModel with thinking and other additional fields."""
    return BedrockModel(
        model_id="anthropic.claude-sonnet-4-20250514-v1:0",
        additional_request_fields={
            "thinking": {"type": "enabled", "budget_tokens": 5000},
            "some_other_field": "value",
        },
    )


def test_thinking_removed_when_forcing_tool_any(model_with_thinking):
    """Thinking should be removed when tool_choice forces tool use with 'any'."""
    tool_choice = {"any": {}}
    result = model_with_thinking._get_additional_request_fields(tool_choice)
    assert result == {}  # thinking removed, no other fields


def test_thinking_removed_when_forcing_specific_tool(model_with_thinking):
    """Thinking should be removed when tool_choice forces a specific tool."""
    tool_choice = {"tool": {"name": "structured_output_tool"}}
    result = model_with_thinking._get_additional_request_fields(tool_choice)
    assert result == {}  # thinking removed, no other fields


def test_thinking_preserved_with_auto_tool_choice(model_with_thinking):
    """Thinking should be preserved when tool_choice is 'auto'."""
    tool_choice = {"auto": {}}
    result = model_with_thinking._get_additional_request_fields(tool_choice)
    assert result == {"additionalModelRequestFields": {"thinking": {"type": "enabled", "budget_tokens": 5000}}}


def test_thinking_preserved_with_none_tool_choice(model_with_thinking):
    """Thinking should be preserved when tool_choice is None."""
    result = model_with_thinking._get_additional_request_fields(None)
    assert result == {"additionalModelRequestFields": {"thinking": {"type": "enabled", "budget_tokens": 5000}}}


def test_other_fields_preserved_when_thinking_removed(model_with_thinking_and_other_fields):
    """Other additional fields should be preserved when thinking is removed."""
    tool_choice = {"any": {}}
    result = model_with_thinking_and_other_fields._get_additional_request_fields(tool_choice)
    assert result == {"additionalModelRequestFields": {"some_other_field": "value"}}


def test_no_fields_when_model_has_no_additional_fields(model_without_thinking):
    """Should return empty dict when model has no additional_request_fields."""
    tool_choice = {"any": {}}
    result = model_without_thinking._get_additional_request_fields(tool_choice)
    assert result == {}


def test_fields_preserved_when_no_thinking_and_forcing_tool():
    """Additional fields without thinking should be preserved when forcing tool."""
    model = BedrockModel(
        model_id="anthropic.claude-sonnet-4-20250514-v1:0",
        additional_request_fields={"some_field": "value"},
    )
    tool_choice = {"any": {}}
    result = model._get_additional_request_fields(tool_choice)
    assert result == {"additionalModelRequestFields": {"some_field": "value"}}



"""Retry strategy implementations for handling model throttling and other retry scenarios.

This module provides hook-based retry strategies that can be configured on the Agent
to control retry behavior for model invocations. Retry strategies implement the
HookProvider protocol and register callbacks for AfterModelCallEvent to determine
when and how to retry failed model calls.
"""

import asyncio
import logging
from typing import Any

from ..hooks.events import AfterInvocationEvent, AfterModelCallEvent
from ..hooks.registry import HookProvider, HookRegistry
from ..types._events import EventLoopThrottleEvent, TypedEvent
from ..types.exceptions import ModelThrottledException

logger = logging.getLogger(__name__)


class ModelRetryStrategy(HookProvider):
    """Default retry strategy for model throttling with exponential backoff.

    Retries model calls on ModelThrottledException using exponential backoff.
    Delay doubles after each attempt: initial_delay, initial_delay*2, initial_delay*4,
    etc., capped at max_delay. State resets after successful calls.

    With defaults (initial_delay=4, max_delay=240, max_attempts=6), delays are:
    4s → 8s → 16s → 32s → 64s (5 retries before giving up on the 6th attempt).

    Args:
        max_attempts: Total model attempts before re-raising the exception.
        initial_delay: Base delay in seconds; used for first two retries, then doubles.
        max_delay: Upper bound in seconds for the exponential backoff.
    """

    def __init__(
        self,
        *,
        max_attempts: int = 6,
        initial_delay: int = 4,
        max_delay: int = 240,
    ):
        """Initialize the retry strategy.

        Args:
            max_attempts: Total model attempts before re-raising the exception. Defaults to 6.
            initial_delay: Base delay in seconds; used for first two retries, then doubles.
                Defaults to 4.
            max_delay: Upper bound in seconds for the exponential backoff. Defaults to 240.
        """
        self._max_attempts = max_attempts
        self._initial_delay = initial_delay
        self._max_delay = max_delay
        self._current_attempt = 0
        self._backwards_compatible_event_to_yield: TypedEvent | None = None

    def register_hooks(self, registry: HookRegistry, **kwargs: Any) -> None:
        """Register callbacks for AfterModelCallEvent and AfterInvocationEvent.

        Args:
            registry: The hook registry to register callbacks with.
            **kwargs: Additional keyword arguments for future extensibility.
        """
        registry.add_callback(AfterModelCallEvent, self._handle_after_model_call)
        registry.add_callback(AfterInvocationEvent, self._handle_after_invocation)

    def _calculate_delay(self, attempt: int) -> int:
        """Calculate retry delay using exponential backoff.

        Args:
            attempt: The attempt number (0-indexed) to calculate delay for.

        Returns:
            Delay in seconds for the given attempt.
        """
        delay: int = self._initial_delay * (2**attempt)
        return min(delay, self._max_delay)

    def _reset_retry_state(self) -> None:
        """Reset retry state to initial values."""
        self._current_attempt = 0

    async def _handle_after_invocation(self, event: AfterInvocationEvent) -> None:
        """Reset retry state after invocation completes.

        Args:
            event: The AfterInvocationEvent signaling invocation completion.
        """
        self._reset_retry_state()

    async def _handle_after_model_call(self, event: AfterModelCallEvent) -> None:
        """Handle model call completion and determine if retry is needed.

        This callback is invoked after each model call. If the call failed with
        a ModelThrottledException and we haven't exceeded max_attempts, it sets
        event.retry to True and sleeps for the current delay before returning.

        On successful calls, it resets the retry state to prepare for future calls.

        Args:
            event: The AfterModelCallEvent containing call results or exception.
        """
        delay = self._calculate_delay(self._current_attempt)

        self._backwards_compatible_event_to_yield = None

        # If already retrying, skip processing (another hook may have triggered retry)
        if event.retry:
            return

        # If model call succeeded, reset retry state
        if event.stop_response is not None:
            logger.debug(
                "stop_reason=<%s> | model call succeeded, resetting retry state",
                event.stop_response.stop_reason,
            )
            self._reset_retry_state()
            return

        # Check if we have an exception and reset state if no exception
        if event.exception is None:
            self._reset_retry_state()
            return

        # Only retry on ModelThrottledException
        if not isinstance(event.exception, ModelThrottledException):
            return

        # Increment attempt counter first
        self._current_attempt += 1

        # Check if we've exceeded max attempts
        if self._current_attempt >= self._max_attempts:
            logger.debug(
                "current_attempt=<%d>, max_attempts=<%d> | max retry attempts reached, not retrying",
                self._current_attempt,
                self._max_attempts,
            )
            return

        self._backwards_compatible_event_to_yield = EventLoopThrottleEvent(delay=delay)

        # Retry the model call
        logger.debug(
            "retry_delay_seconds=<%s>, max_attempts=<%s>, current_attempt=<%s> "
            "| throttling exception encountered | delaying before next retry",
            delay,
            self._max_attempts,
            self._current_attempt,
        )

        # Sleep for current delay
        await asyncio.sleep(delay)

        # Set retry flag and track that this strategy triggered it
        event.retry = True



import pytest

from strands.hooks import AfterToolCallEvent, BeforeToolCallEvent
from strands.interrupt import Interrupt
from strands.tools.executors import ConcurrentToolExecutor
from strands.tools.structured_output._structured_output_context import StructuredOutputContext
from strands.types._events import ToolInterruptEvent, ToolResultEvent


@pytest.fixture
def executor():
    return ConcurrentToolExecutor()


@pytest.fixture
def structured_output_context():
    return StructuredOutputContext(structured_output_model=None)


@pytest.mark.asyncio
async def test_concurrent_executor_execute(
    executor, agent, tool_results, cycle_trace, cycle_span, invocation_state, structured_output_context, alist
):
    tool_uses = [
        {"name": "weather_tool", "toolUseId": "1", "input": {}},
        {"name": "temperature_tool", "toolUseId": "2", "input": {}},
    ]
    stream = executor._execute(
        agent, tool_uses, tool_results, cycle_trace, cycle_span, invocation_state, structured_output_context
    )

    tru_events = sorted(await alist(stream), key=lambda event: event.tool_use_id)
    exp_events = [
        ToolResultEvent({"toolUseId": "1", "status": "success", "content": [{"text": "sunny"}]}),
        ToolResultEvent({"toolUseId": "2", "status": "success", "content": [{"text": "75F"}]}),
    ]
    assert tru_events == exp_events

    tru_results = sorted(tool_results, key=lambda result: result.get("toolUseId"))
    exp_results = [exp_events[0].tool_result, exp_events[1].tool_result]
    assert tru_results == exp_results


@pytest.mark.asyncio
async def test_concurrent_executor_interrupt(
    executor, agent, tool_results, cycle_trace, cycle_span, invocation_state, structured_output_context, alist
):
    interrupt = Interrupt(
        id="v1:before_tool_call:test_tool_id_1:78714d6c-613c-5cf4-bf25-7037569941f9",
        name="test_name",
        reason="test reason",
    )

    def interrupt_callback(event):
        if event.tool_use["name"] == "weather_tool":
            event.interrupt("test_name", "test reason")

    agent.hooks.add_callback(BeforeToolCallEvent, interrupt_callback)

    tool_uses = [
        {"name": "weather_tool", "toolUseId": "test_tool_id_1", "input": {}},
        {"name": "temperature_tool", "toolUseId": "test_tool_id_2", "input": {}},
    ]

    stream = executor._execute(
        agent, tool_uses, tool_results, cycle_trace, cycle_span, invocation_state, structured_output_context
    )

    tru_events = sorted(await alist(stream), key=lambda event: event.tool_use_id)
    exp_events = [
        ToolInterruptEvent(tool_uses[0], [interrupt]),
        ToolResultEvent({"toolUseId": "test_tool_id_2", "status": "success", "content": [{"text": "75F"}]}),
    ]
    assert tru_events == exp_events

    tru_results = tool_results
    exp_results = [exp_events[1].tool_result]
    assert tru_results == exp_results


@pytest.mark.asyncio
async def test_concurrent_executor_reraises_exceptions(
    executor, agent, tool_results, cycle_trace, cycle_span, invocation_state, structured_output_context, alist
):
    """Test that hook re-raised exceptions propagate and cancel remaining tasks."""

    def reraise_callback(event):
        if event.exception is not None:
            raise event.exception

    agent.hooks.add_callback(AfterToolCallEvent, reraise_callback)

    tool_uses = [
        {"name": "exception_tool", "toolUseId": "1", "input": {}},
        {"name": "slow_tool", "toolUseId": "2", "input": {}},
    ]

    stream = executor._execute(
        agent, tool_uses, tool_results, cycle_trace, cycle_span, invocation_state, structured_output_context
    )

    with pytest.raises(RuntimeError, match="Tool error"):
        await alist(stream)

    assert tool_results == []



"""Type definitions for bidirectional streaming."""

from .agent import BidiAgentInput
from .events import (
    BidiAudioInputEvent,
    BidiAudioStreamEvent,
    BidiConnectionCloseEvent,
    BidiConnectionRestartEvent,
    BidiConnectionStartEvent,
    BidiErrorEvent,
    BidiImageInputEvent,
    BidiInputEvent,
    BidiInterruptionEvent,
    BidiOutputEvent,
    BidiResponseCompleteEvent,
    BidiResponseStartEvent,
    BidiTextInputEvent,
    BidiTranscriptStreamEvent,
    BidiUsageEvent,
    ModalityUsage,
)
from .io import BidiInput, BidiOutput

__all__ = [
    "BidiInput",
    "BidiOutput",
    "BidiAgentInput",
    # Input Events
    "BidiTextInputEvent",
    "BidiAudioInputEvent",
    "BidiImageInputEvent",
    "BidiInputEvent",
    # Output Events
    "BidiConnectionStartEvent",
    "BidiConnectionRestartEvent",
    "BidiConnectionCloseEvent",
    "BidiResponseStartEvent",
    "BidiResponseCompleteEvent",
    "BidiAudioStreamEvent",
    "BidiTranscriptStreamEvent",
    "BidiInterruptionEvent",
    "BidiUsageEvent",
    "ModalityUsage",
    "BidiErrorEvent",
    "BidiOutputEvent",
]



"""SDK model providers.

This package includes an abstract base Model class along with concrete implementations for specific providers.
"""

from typing import Any

from . import bedrock, model
from .bedrock import BedrockModel
from .model import CacheConfig, Model

__all__ = [
    "bedrock",
    "model",
    "BedrockModel",
    "CacheConfig",
    "Model",
]


def __getattr__(name: str) -> Any:
    """Lazy load model implementations only when accessed.

    This defers the import of optional dependencies until actually needed.
    """
    if name == "AnthropicModel":
        from .anthropic import AnthropicModel

        return AnthropicModel
    if name == "GeminiModel":
        from .gemini import GeminiModel

        return GeminiModel
    if name == "LiteLLMModel":
        from .litellm import LiteLLMModel

        return LiteLLMModel
    if name == "LlamaAPIModel":
        from .llamaapi import LlamaAPIModel

        return LlamaAPIModel
    if name == "LlamaCppModel":
        from .llamacpp import LlamaCppModel

        return LlamaCppModel
    if name == "MistralModel":
        from .mistral import MistralModel

        return MistralModel
    if name == "OllamaModel":
        from .ollama import OllamaModel

        return OllamaModel
    if name == "OpenAIModel":
        from .openai import OpenAIModel

        return OpenAIModel
    if name == "OpenAIResponsesModel":
        from .openai_responses import OpenAIResponsesModel

        return OpenAIResponsesModel
    if name == "SageMakerAIModel":
        from .sagemaker import SageMakerAIModel

        return SageMakerAIModel
    if name == "WriterModel":
        from .writer import WriterModel

        return WriterModel
    raise AttributeError(f"cannot import name '{name}' from '{__name__}' ({__file__})")



"""Additional A2A types."""

from typing import Any, TypeAlias

from a2a.types import Message, Task, TaskArtifactUpdateEvent, TaskStatusUpdateEvent

from ._events import TypedEvent

A2AResponse: TypeAlias = tuple[Task, TaskStatusUpdateEvent | TaskArtifactUpdateEvent | None] | Message | Any


class A2AStreamEvent(TypedEvent):
    """Event emitted for every update received from the remote A2A server.

    This event wraps all A2A response types during streaming, including:
    - Partial task updates (TaskArtifactUpdateEvent)
    - Status updates (TaskStatusUpdateEvent)
    - Complete messages (Message)
    - Final task completions

    The event is emitted for EVERY update from the server, regardless of whether
    it represents a complete or partial response. When streaming completes, an
    AgentResultEvent containing the final AgentResult is also emitted after all
    A2AStreamEvents.
    """

    def __init__(self, a2a_event: A2AResponse) -> None:
        """Initialize with A2A event.

        Args:
            a2a_event: The original A2A event (Task tuple or Message)
        """
        super().__init__(
            {
                "type": "a2a_stream",
                "event": a2a_event,  # Nest A2A event to avoid field conflicts
            }
        )



"""MCP server with task-augmented tool execution support for integration testing."""

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from typing import Any

import click
import mcp.types as types
from mcp.server.experimental.task_context import ServerTaskContext
from mcp.server.lowlevel import Server
from mcp.server.streamable_http_manager import StreamableHTTPSessionManager
from starlette.applications import Starlette
from starlette.routing import Mount


def create_task_server() -> Server:
    """Create and configure the task-supporting MCP server."""
    server = Server("task-echo-server")
    server.experimental.enable_tasks()

    # Workaround: MCP Python SDK's enable_tasks() doesn't properly set tasks.requests.tools.call capability
    original_update_capabilities = server.experimental.update_capabilities

    def patched_update_capabilities(capabilities: types.ServerCapabilities) -> None:
        original_update_capabilities(capabilities)
        if capabilities.tasks and capabilities.tasks.requests and capabilities.tasks.requests.tools:
            capabilities.tasks.requests.tools.call = types.TasksCallCapability()

    server.experimental.update_capabilities = patched_update_capabilities  # type: ignore[method-assign]

    @server.list_tools()
    async def list_tools() -> list[types.Tool]:
        return [
            types.Tool(
                name="task_required_echo",
                description="Echo that requires task-augmented execution",
                inputSchema={"type": "object", "properties": {"message": {"type": "string"}}, "required": ["message"]},
                execution=types.ToolExecution(taskSupport=types.TASK_REQUIRED),
            ),
            types.Tool(
                name="task_optional_echo",
                description="Echo that optionally supports task-augmented execution",
                inputSchema={"type": "object", "properties": {"message": {"type": "string"}}, "required": ["message"]},
                execution=types.ToolExecution(taskSupport=types.TASK_OPTIONAL),
            ),
            types.Tool(
                name="task_forbidden_echo",
                description="Echo that does not support task-augmented execution",
                inputSchema={"type": "object", "properties": {"message": {"type": "string"}}, "required": ["message"]},
                execution=types.ToolExecution(taskSupport=types.TASK_FORBIDDEN),
            ),
            types.Tool(
                name="echo",
                description="Simple echo without task support setting",
                inputSchema={"type": "object", "properties": {"message": {"type": "string"}}, "required": ["message"]},
            ),
        ]

    async def handle_task_required_echo(arguments: dict[str, Any]) -> types.CreateTaskResult:
        ctx = server.request_context
        ctx.experimental.validate_task_mode(types.TASK_REQUIRED)
        message = arguments.get("message", "")

        async def work(task: ServerTaskContext) -> types.CallToolResult:
            await task.update_status("Processing echo...")
            return types.CallToolResult(content=[types.TextContent(type="text", text=f"Task echo: {message}")])

        return await ctx.experimental.run_task(work)

    async def handle_task_optional_echo(arguments: dict[str, Any]) -> types.CallToolResult | types.CreateTaskResult:
        ctx = server.request_context
        message = arguments.get("message", "")

        if ctx.experimental.is_task:

            async def work(task: ServerTaskContext) -> types.CallToolResult:
                await task.update_status("Processing optional task echo...")
                return types.CallToolResult(
                    content=[types.TextContent(type="text", text=f"Task optional echo: {message}")]
                )

            return await ctx.experimental.run_task(work)
        else:
            return types.CallToolResult(
                content=[types.TextContent(type="text", text=f"Direct optional echo: {message}")]
            )

    async def handle_task_forbidden_echo(arguments: dict[str, Any]) -> types.CallToolResult:
        message = arguments.get("message", "")
        return types.CallToolResult(content=[types.TextContent(type="text", text=f"Forbidden echo: {message}")])

    async def handle_simple_echo(arguments: dict[str, Any]) -> types.CallToolResult:
        message = arguments.get("message", "")
        return types.CallToolResult(content=[types.TextContent(type="text", text=f"Simple echo: {message}")])

    @server.call_tool()
    async def handle_call_tool(name: str, arguments: dict[str, Any]) -> types.CallToolResult | types.CreateTaskResult:
        handlers = {
            "task_required_echo": handle_task_required_echo,
            "task_optional_echo": handle_task_optional_echo,
            "task_forbidden_echo": handle_task_forbidden_echo,
            "echo": handle_simple_echo,
        }
        if name in handlers:
            return await handlers[name](arguments)
        return types.CallToolResult(
            content=[types.TextContent(type="text", text=f"Unknown tool: {name}")], isError=True
        )

    return server


def create_starlette_app(port: int) -> tuple[Starlette, StreamableHTTPSessionManager]:
    """Create the Starlette app with MCP session manager."""
    server = create_task_server()
    session_manager = StreamableHTTPSessionManager(app=server)

    @asynccontextmanager
    async def app_lifespan(app: Starlette) -> AsyncIterator[None]:
        async with session_manager.run():
            yield

    return Starlette(routes=[Mount("/mcp", app=session_manager.handle_request)], lifespan=app_lifespan), session_manager


@click.command()
@click.option("--port", default=8010, help="Port to listen on")
def main(port: int) -> int:
    """Start the task echo server."""
    import uvicorn

    starlette_app, _ = create_starlette_app(port)
    print(f"Starting task echo server on http://localhost:{port}/mcp")
    uvicorn.run(starlette_app, host="127.0.0.1", port=port)
    return 0


if __name__ == "__main__":
    main()



"""Integration tests for agent cancellation with Amazon Bedrock.

These tests verify that cancellation works correctly with the Bedrock model provider.
They require valid AWS credentials and may incur API costs.

To run these tests:
    hatch run test-integ tests_integ/test_cancellation.py
"""

import asyncio
import os
import threading

import pytest

from strands import Agent, tool
from strands.hooks import AfterModelCallEvent, BeforeModelCallEvent
from strands.models import BedrockModel

# Skip all tests if no AWS credentials are available
pytestmark = [
    pytest.mark.skipif(not os.getenv("AWS_REGION"), reason="AWS credentials not available"),
    pytest.mark.asyncio,
]


async def test_cancel_with_bedrock():
    """Test agent.cancel() with Amazon Bedrock model.

    Verifies that cancellation works correctly with a real Bedrock
    model by cancelling before the model call starts.
    """

    agent = Agent(model=BedrockModel(model_id="anthropic.claude-3-haiku-20240307-v1:0"))

    # Cancel deterministically before the model call
    async def cancel_before_model(event: BeforeModelCallEvent):
        agent.cancel()

    agent.add_hook(cancel_before_model, BeforeModelCallEvent)

    result = await agent.invoke_async(
        "Write a detailed 1000-word essay about the history of space exploration, "
        "including major milestones, key figures, and technological breakthroughs."
    )

    assert result.stop_reason == "cancelled"
    assert result.message["role"] == "assistant"
    assert result.message["content"] == [{"text": "Cancelled by user"}]


async def test_cancel_during_streaming_bedrock():
    """Test agent.cancel() during streaming with Bedrock.

    Verifies that cancellation works correctly when using the
    streaming API with a real Bedrock model.
    """

    agent = Agent(model=BedrockModel(model_id="anthropic.claude-3-haiku-20240307-v1:0"))

    events = []
    async for event in agent.stream_async(
        "Write a detailed story about a space adventure. Make it at least 500 words long."
    ):
        events.append(event)
        # Cancel after receiving the first model delta event
        if "data" in event:
            agent.cancel()
        if event.get("result"):
            break

    # Find the result event
    result_event = next((e for e in events if e.get("result")), None)
    assert result_event is not None
    assert result_event["result"].stop_reason == "cancelled"


async def test_cancel_with_tools_bedrock():
    """Test agent.cancel() during tool execution with Bedrock.

    Verifies that cancellation works correctly when the agent
    is executing tools with a real Bedrock model.
    """

    @tool
    async def slow_calculation(x: int, y: int) -> int:
        """Perform a slow calculation that takes time.

        Args:
            x: First number
            y: Second number

        Returns:
            The sum of x and y
        """
        await asyncio.sleep(2)
        return x + y

    @tool
    async def another_calculation(a: int, b: int) -> int:
        """Another slow calculation.

        Args:
            a: First number
            b: Second number

        Returns:
            The product of a and b
        """
        await asyncio.sleep(2)
        return a * b

    agent = Agent(
        model=BedrockModel(model_id="anthropic.claude-3-haiku-20240307-v1:0"),
        tools=[slow_calculation, another_calculation],
    )

    # Cancel deterministically after model returns tool_use
    async def cancel_after_model(event: AfterModelCallEvent):
        if event.stop_response and event.stop_response.stop_reason == "tool_use":
            agent.cancel()

    agent.add_hook(cancel_after_model, AfterModelCallEvent)

    result = await agent.invoke_async(
        "Please use the slow_calculation tool to add 5 and 10, then use another_calculation to multiply 3 and 7."
    )

    assert result.stop_reason == "cancelled"


async def test_cancel_from_thread_bedrock():
    """Test agent.cancel() from a different thread with Bedrock.

    Simulates a real-world scenario where cancellation is triggered
    from a different thread (e.g., a web request handler) while the agent
    is executing.
    """

    agent = Agent(model=BedrockModel(model_id="anthropic.claude-3-haiku-20240307-v1:0"))

    # Cancel deterministically from a different thread before the model call
    def cancel_before_model(event: BeforeModelCallEvent):
        thread = threading.Thread(target=agent.cancel)
        thread.start()
        thread.join()

    agent.add_hook(cancel_before_model, BeforeModelCallEvent)

    result = await agent.invoke_async(
        "Write a comprehensive guide about machine learning, "
        "covering supervised learning, unsupervised learning, and deep learning. "
        "Make it at least 800 words."
    )

    assert result.stop_reason == "cancelled"



"""LLM-based steering handler that uses an LLM to provide contextual guidance."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Literal, cast

from pydantic import BaseModel, Field

from .....models import Model
from .....types.tools import ToolUse
from ...context_providers.ledger_provider import LedgerProvider
from ...core.action import Guide, Interrupt, Proceed, ToolSteeringAction
from ...core.context import SteeringContextProvider
from ...core.handler import SteeringHandler
from .mappers import DefaultPromptMapper, LLMPromptMapper

if TYPE_CHECKING:
    from .....agent import Agent

logger = logging.getLogger(__name__)


class _LLMSteering(BaseModel):
    """Structured output model for LLM steering decisions."""

    decision: Literal["proceed", "guide", "interrupt"] = Field(
        description="Steering decision: 'proceed' to continue, 'guide' to provide feedback, 'interrupt' for human input"
    )
    reason: str = Field(description="Clear explanation of the steering decision and any guidance provided")


class LLMSteeringHandler(SteeringHandler):
    """Steering handler that uses an LLM to provide contextual guidance.

    Uses natural language prompts to evaluate tool calls and provide
    contextual steering guidance to help agents navigate complex workflows.
    """

    def __init__(
        self,
        system_prompt: str,
        prompt_mapper: LLMPromptMapper | None = None,
        model: Model | None = None,
        context_providers: list[SteeringContextProvider] | None = None,
    ):
        """Initialize the LLMSteeringHandler.

        Args:
            system_prompt: System prompt defining steering guidance rules
            prompt_mapper: Custom prompt mapper for evaluation prompts
            model: Optional model override for steering evaluation
            context_providers: List of context providers for populating steering context.
                Defaults to [LedgerProvider()] if None. Pass an empty list to disable
                context providers.
        """
        providers: list[SteeringContextProvider] = (
            [LedgerProvider()] if context_providers is None else context_providers
        )
        super().__init__(context_providers=providers)
        self.system_prompt = system_prompt
        self.prompt_mapper = prompt_mapper or DefaultPromptMapper()
        self.model = model

    async def steer_before_tool(self, *, agent: Agent, tool_use: ToolUse, **kwargs: Any) -> ToolSteeringAction:
        """Provide contextual guidance for tool usage.

        Args:
            agent: The agent instance
            tool_use: The tool use object with name and arguments
            **kwargs: Additional keyword arguments for steering evaluation

        Returns:
            SteeringAction indicating how to guide the tool execution
        """
        # Generate steering prompt
        prompt = self.prompt_mapper.create_steering_prompt(self.steering_context, tool_use=tool_use)

        # Create isolated agent for steering evaluation (no shared conversation state)
        from .....agent import Agent

        steering_agent = Agent(system_prompt=self.system_prompt, model=self.model or agent.model, callback_handler=None)

        # Get LLM decision
        llm_result: _LLMSteering = cast(
            _LLMSteering, steering_agent(prompt, structured_output_model=_LLMSteering).structured_output
        )

        # Convert LLM decision to steering action
        match llm_result.decision:
            case "proceed":
                return Proceed(reason=llm_result.reason)
            case "guide":
                return Guide(reason=llm_result.reason)
            case "interrupt":
                return Interrupt(reason=llm_result.reason)
            case _:
                logger.warning("decision=<%s> | unknown llm decision, defaulting to proceed", llm_result.decision)  # type: ignore[unreachable]
                return Proceed(reason="Unknown LLM decision, defaulting to proceed")



"""Deprecated: Use strands.vended_plugins.steering.core.context instead."""

import warnings
from typing import Any

_TARGET_MODULE = "strands.vended_plugins.steering.core.context"


def __getattr__(name: str) -> Any:
    from strands.vended_plugins.steering.core import context

    obj = getattr(context, name, None)
    if obj is not None:
        warnings.warn(
            f"{name} has been moved to production. Use {name} from {_TARGET_MODULE} instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return obj
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


__all__: list[str] = []



"""Unit tests for Nova Sonic bidirectional model implementation.

Tests the unified BidirectionalModel interface implementation for Amazon Nova Sonic,
covering connection lifecycle, event conversion, audio streaming, and tool execution.
"""

import sys

if sys.version_info < (3, 12):
    import pytest

    pytest.skip(reason="BidiNovaSonicModel is only supported for Python 3.12+", allow_module_level=True)

import asyncio
import base64
import json
from unittest.mock import AsyncMock, Mock, patch

import pytest
import pytest_asyncio
from aws_sdk_bedrock_runtime.models import ModelTimeoutException, ValidationException

from strands.experimental.bidi.models.model import BidiModelTimeoutError
from strands.experimental.bidi.models.nova_sonic import (
    BidiNovaSonicModel,
    NOVA_SONIC_V1_MODEL_ID,
    NOVA_SONIC_V2_MODEL_ID,
)
from strands.experimental.bidi.types.events import (
    BidiAudioInputEvent,
    BidiAudioStreamEvent,
    BidiImageInputEvent,
    BidiInterruptionEvent,
    BidiResponseStartEvent,
    BidiTextInputEvent,
    BidiTranscriptStreamEvent,
    BidiUsageEvent,
)
from strands.types._events import ToolResultEvent
from strands.types.tools import ToolResult


# Test fixtures
@pytest.fixture
def model_id():
    """Nova Sonic model identifier."""
    return "amazon.nova-sonic-v1:0"


@pytest.fixture
def boto_session():
    return Mock(region_name="us-east-1")


@pytest.fixture
def mock_stream():
    """Mock Nova Sonic bidirectional stream."""
    stream = AsyncMock()
    stream.input_stream = AsyncMock()
    stream.input_stream.send = AsyncMock()
    stream.input_stream.close = AsyncMock()
    stream.await_output = AsyncMock()
    return stream


@pytest.fixture
def mock_client(mock_stream):
    """Mock Bedrock Runtime client."""
    with patch("strands.experimental.bidi.models.nova_sonic.BedrockRuntimeClient") as mock_cls:
        mock_instance = AsyncMock()
        mock_instance.invoke_model_with_bidirectional_stream = AsyncMock(return_value=mock_stream)
        mock_cls.return_value = mock_instance

        yield mock_instance


@pytest_asyncio.fixture
def nova_model(model_id, boto_session, mock_client):
    """Create Nova Sonic model instance."""
    _ = mock_client

    model = BidiNovaSonicModel(model_id=model_id, client_config={"boto_session": boto_session})
    yield model


# Initialization and Connection Tests


@pytest.mark.asyncio
async def test_model_initialization(model_id, boto_session):
    """Test model initialization with configuration."""
    model = BidiNovaSonicModel(model_id=model_id, client_config={"boto_session": boto_session})

    assert model.model_id == model_id
    assert model.region == "us-east-1"
    assert model._connection_id is None


# Audio Configuration Tests


@pytest.mark.asyncio
async def test_audio_config_defaults(model_id, boto_session):
    """Test default audio configuration."""
    model = BidiNovaSonicModel(model_id=model_id, client_config={"boto_session": boto_session})

    assert model.config["audio"]["input_rate"] == 16000
    assert model.config["audio"]["output_rate"] == 16000
    assert model.config["audio"]["channels"] == 1
    assert model.config["audio"]["format"] == "pcm"
    assert model.config["audio"]["voice"] == "matthew"


@pytest.mark.asyncio
async def test_audio_config_partial_override(model_id, boto_session):
    """Test partial audio configuration override."""
    provider_config = {"audio": {"output_rate": 24000, "voice": "ruth"}}
    model = BidiNovaSonicModel(
        model_id=model_id, client_config={"boto_session": boto_session}, provider_config=provider_config
    )

    # Overridden values
    assert model.config["audio"]["output_rate"] == 24000
    assert model.config["audio"]["voice"] == "ruth"

    # Default values preserved
    assert model.config["audio"]["input_rate"] == 16000
    assert model.config["audio"]["channels"] == 1
    assert model.config["audio"]["format"] == "pcm"


@pytest.mark.asyncio
async def test_audio_config_full_override(model_id, boto_session):
    """Test full audio configuration override."""
    provider_config = {
        "audio": {
            "input_rate": 48000,
            "output_rate": 48000,
            "channels": 2,
            "format": "pcm",
            "voice": "stephen",
        }
    }
    model = BidiNovaSonicModel(
        model_id=model_id, client_config={"boto_session": boto_session}, provider_config=provider_config
    )

    assert model.config["audio"]["input_rate"] == 48000
    assert model.config["audio"]["output_rate"] == 48000
    assert model.config["audio"]["channels"] == 2
    assert model.config["audio"]["format"] == "pcm"
    assert model.config["audio"]["voice"] == "stephen"


@pytest.mark.asyncio
async def test_connection_lifecycle(nova_model, mock_client, mock_stream):
    """Test complete connection lifecycle with various configurations."""

    # Test basic connection
    await nova_model.start(system_prompt="Test system prompt")
    assert nova_model._stream == mock_stream
    assert nova_model._connection_id is not None
    assert mock_client.invoke_model_with_bidirectional_stream.called

    # Test close
    await nova_model.stop()
    assert mock_stream.close.called

    # Test connection with tools
    tools = [
        {
            "name": "get_weather",
            "description": "Get weather information",
            "inputSchema": {"json": json.dumps({"type": "object", "properties": {}})},
        }
    ]
    await nova_model.start(system_prompt="You are helpful", tools=tools)
    # Verify initialization events were sent (connectionStart, promptStart, system prompt)
    assert mock_stream.input_stream.send.call_count >= 3
    await nova_model.stop()


@pytest.mark.asyncio
async def test_model_stop_alone(nova_model):
    await nova_model.stop()  # Should not raise


@pytest.mark.asyncio
async def test_connection_with_message_history(nova_model, mock_client, mock_stream):
    """Test connection initialization with conversation history."""
    nova_model.client = mock_client

    # Create message history
    messages = [
        {"role": "user", "content": [{"text": "What's the weather?"}]},
        {"role": "assistant", "content": [{"text": "I'll check the weather for you."}]},
        {
            "role": "assistant",
            "content": [{"toolUse": {"toolUseId": "tool-123", "name": "get_weather", "input": {}}}],
        },
        {
            "role": "user",
            "content": [{"toolResult": {"toolUseId": "tool-123", "content": [{"text": "Sunny, 72°F"}]}}],
        },
        {"role": "assistant", "content": [{"text": "It's sunny and 72 degrees."}]},
    ]

    # Start connection with message history
    await nova_model.start(system_prompt="You are a helpful assistant", messages=messages)

    # Verify initialization events were sent
    # Should include: sessionStart, promptStart, system prompt (3 events),
    # and message history (only text messages: 3 messages * 3 events each = 9 events)
    # Tool use/result messages are now skipped in history
    # Total: 1 + 1 + 3 + 9 = 14 events minimum
    assert mock_stream.input_stream.send.call_count >= 14

    # Verify the events contain proper role information
    sent_events = [call.args[0].value.bytes_.decode("utf-8") for call in mock_stream.input_stream.send.call_args_list]

    # Check that USER and ASSISTANT roles are present in contentStart events
    user_events = [e for e in sent_events if '"role": "USER"' in e]
    assistant_events = [e for e in sent_events if '"role": "ASSISTANT"' in e]

    # Only text messages are sent, so we expect 1 user message and 2 assistant messages
    assert len(user_events) >= 1
    assert len(assistant_events) >= 2

    await nova_model.stop()


# Send Method Tests


@pytest.mark.asyncio
async def test_send_all_content_types(nova_model, mock_stream):
    """Test sending all content types through unified send() method."""
    await nova_model.start()

    # Test text content
    text_event = BidiTextInputEvent(text="Hello, Nova!", role="user")
    await nova_model.send(text_event)
    # Should send contentStart, textInput, and contentEnd
    assert mock_stream.input_stream.send.call_count >= 3

    # Test audio content (base64 encoded)
    audio_b64 = base64.b64encode(b"audio data").decode("utf-8")
    audio_event = BidiAudioInputEvent(audio=audio_b64, format="pcm", sample_rate=16000, channels=1)
    await nova_model.send(audio_event)
    # Should start audio connection and send audio
    assert nova_model._audio_content_name
    assert mock_stream.input_stream.send.called

    # Test tool result with single content item (should be unwrapped)
    tool_result_single: ToolResult = {
        "toolUseId": "tool-123",
        "status": "success",
        "content": [{"text": "Weather is sunny"}],
    }
    await nova_model.send(ToolResultEvent(tool_result_single))
    # Should send contentStart, toolResult, and contentEnd
    assert mock_stream.input_stream.send.called

    # Test tool result with multiple content items (should send as array)
    tool_result_multi: ToolResult = {
        "toolUseId": "tool-456",
        "status": "success",
        "content": [{"text": "Part 1"}, {"json": {"data": "value"}}],
    }
    await nova_model.send(ToolResultEvent(tool_result_multi))
    assert mock_stream.input_stream.send.called

    await nova_model.stop()


@pytest.mark.asyncio
async def test_send_edge_cases(nova_model):
    """Test send() edge cases and error handling."""

    # Test image content (not supported, base64 encoded, no encoding parameter)
    await nova_model.start()
    image_b64 = base64.b64encode(b"image data").decode("utf-8")
    image_event = BidiImageInputEvent(
        image=image_b64,
        mime_type="image/jpeg",
    )

    with pytest.raises(ValueError, match=r"content not supported"):
        await nova_model.send(image_event)

    await nova_model.stop()


# Receive and Event Conversion Tests


@pytest.mark.asyncio
async def test_event_conversion(nova_model):
    """Test conversion of all Nova Sonic event types to standard format."""
    # Test audio output (now returns BidiAudioStreamEvent)
    audio_bytes = b"test audio data"
    audio_base64 = base64.b64encode(audio_bytes).decode("utf-8")
    nova_event = {"audioOutput": {"content": audio_base64}}
    result = nova_model._convert_nova_event(nova_event)
    assert result is not None
    assert isinstance(result, BidiAudioStreamEvent)
    assert result.get("type") == "bidi_audio_stream"
    # Audio is kept as base64 string
    assert result.get("audio") == audio_base64
    assert result.get("format") == "pcm"
    assert result.get("sample_rate") == 16000

    # Test text output (now returns BidiTranscriptStreamEvent)
    nova_event = {"textOutput": {"content": "Hello, world!", "role": "ASSISTANT"}}
    result = nova_model._convert_nova_event(nova_event)
    assert result is not None
    assert isinstance(result, BidiTranscriptStreamEvent)
    assert result.get("type") == "bidi_transcript_stream"
    assert result.get("text") == "Hello, world!"
    assert result.get("role") == "assistant"
    assert result.delta == {"text": "Hello, world!"}
    assert result.current_transcript == "Hello, world!"

    # Test tool use (now returns ToolUseStreamEvent from core strands)
    tool_input = {"location": "Seattle"}
    nova_event = {"toolUse": {"toolUseId": "tool-123", "toolName": "get_weather", "content": json.dumps(tool_input)}}
    result = nova_model._convert_nova_event(nova_event)
    assert result is not None
    # ToolUseStreamEvent has delta and current_tool_use, not a "type" field
    assert "delta" in result
    assert "toolUse" in result["delta"]
    tool_use = result["delta"]["toolUse"]
    assert tool_use["toolUseId"] == "tool-123"
    assert tool_use["name"] == "get_weather"
    assert tool_use["input"] == tool_input

    # Test interruption (now returns BidiInterruptionEvent)
    nova_event = {"stopReason": "INTERRUPTED"}
    result = nova_model._convert_nova_event(nova_event)
    assert result is not None
    assert isinstance(result, BidiInterruptionEvent)
    assert result.get("type") == "bidi_interruption"
    assert result.get("reason") == "user_speech"

    # Test usage metrics (now returns BidiUsageEvent)
    nova_event = {
        "usageEvent": {
            "totalTokens": 100,
            "totalInputTokens": 40,
            "totalOutputTokens": 60,
            "details": {"total": {"output": {"speechTokens": 30}}},
        }
    }
    result = nova_model._convert_nova_event(nova_event)
    assert result is not None
    assert isinstance(result, BidiUsageEvent)
    assert result.get("type") == "bidi_usage"
    assert result.get("totalTokens") == 100
    assert result.get("inputTokens") == 40
    assert result.get("outputTokens") == 60

    # Test content start tracks role and emits BidiResponseStartEvent
    # TEXT type contentStart (matches API spec)
    nova_event = {
        "contentStart": {
            "role": "ASSISTANT",
            "type": "TEXT",
            "additionalModelFields": '{"generationStage":"FINAL"}',
            "contentId": "content-123",
        }
    }
    result = nova_model._convert_nova_event(nova_event)
    assert result is not None
    assert isinstance(result, BidiResponseStartEvent)
    assert result.get("type") == "bidi_response_start"
    assert nova_model._generation_stage == "FINAL"

    # Test AUDIO type contentStart (no additionalModelFields)
    nova_event = {"contentStart": {"role": "ASSISTANT", "type": "AUDIO", "contentId": "content-456"}}
    result = nova_model._convert_nova_event(nova_event)
    assert result is not None
    assert isinstance(result, BidiResponseStartEvent)

    # Test TOOL type contentStart
    nova_event = {"contentStart": {"role": "TOOL", "type": "TOOL", "contentId": "content-789"}}
    result = nova_model._convert_nova_event(nova_event)
    assert result is not None
    assert isinstance(result, BidiResponseStartEvent)


# Audio Streaming Tests


@pytest.mark.asyncio
async def test_audio_connection_lifecycle(nova_model):
    """Test audio connection start and end lifecycle."""

    await nova_model.start()

    # Start audio connection
    await nova_model._start_audio_connection()
    assert nova_model._audio_content_name

    # End audio connection
    await nova_model._end_audio_input()
    assert not nova_model._audio_content_name

    await nova_model.stop()


# Helper Method Tests


@pytest.mark.asyncio
async def test_tool_configuration(nova_model):
    """Test building tool configuration from tool specs."""
    tools = [
        {
            "name": "get_weather",
            "description": "Get weather information",
            "inputSchema": {"json": json.dumps({"type": "object", "properties": {"location": {"type": "string"}}})},
        }
    ]

    tool_config = nova_model._build_tool_configuration(tools)

    assert len(tool_config) == 1
    assert tool_config[0]["toolSpec"]["name"] == "get_weather"
    assert tool_config[0]["toolSpec"]["description"] == "Get weather information"
    assert "inputSchema" in tool_config[0]["toolSpec"]


@pytest.mark.asyncio
async def test_event_templates(nova_model):
    """Test event template generation."""
    # Test connection start event
    event_json = nova_model._get_connection_start_event()
    event = json.loads(event_json)
    assert "event" in event
    assert "sessionStart" in event["event"]
    assert "inferenceConfiguration" in event["event"]["sessionStart"]

    # Test prompt start event
    nova_model._connection_id = "test-connection"
    event_json = nova_model._get_prompt_start_event([])
    event = json.loads(event_json)
    assert "event" in event
    assert "promptStart" in event["event"]
    assert event["event"]["promptStart"]["promptName"] == "test-connection"

    # Test text input event
    content_name = "test-content"
    event_json = nova_model._get_text_input_event(content_name, "Hello")
    event = json.loads(event_json)
    assert "event" in event
    assert "textInput" in event["event"]
    assert event["event"]["textInput"]["content"] == "Hello"

    # Test tool result event
    result = {"result": "Success"}
    event_json = nova_model._get_tool_result_event(content_name, result)
    event = json.loads(event_json)
    assert "event" in event
    assert "toolResult" in event["event"]
    assert json.loads(event["event"]["toolResult"]["content"]) == result


@pytest.mark.asyncio
async def test_message_history_conversion(nova_model):
    """Test conversion of agent messages to Nova Sonic history events."""
    nova_model.connection_id = "test-connection"

    # Test with various message types
    messages = [
        {"role": "user", "content": [{"text": "Hello"}]},
        {"role": "assistant", "content": [{"text": "Hi there!"}]},
        {
            "role": "assistant",
            "content": [{"toolUse": {"toolUseId": "tool-1", "name": "calculator", "input": {"expr": "2+2"}}}],
        },
        {"role": "user", "content": [{"toolResult": {"toolUseId": "tool-1", "content": [{"text": "4"}]}}]},
        {"role": "assistant", "content": [{"text": "The answer is 4"}]},
    ]

    events = nova_model._get_message_history_events(messages)

    # Only text messages generate events (3 messages * 3 events each = 9 events)
    # Tool use/result messages are now skipped in history
    assert len(events) == 9

    # Parse and verify events
    parsed_events = [json.loads(e) for e in events]

    # Check first message (user)
    assert "contentStart" in parsed_events[0]["event"]
    assert parsed_events[0]["event"]["contentStart"]["role"] == "USER"
    assert "textInput" in parsed_events[1]["event"]
    assert parsed_events[1]["event"]["textInput"]["content"] == "Hello"
    assert "contentEnd" in parsed_events[2]["event"]

    # Check second message (assistant)
    assert "contentStart" in parsed_events[3]["event"]
    assert parsed_events[3]["event"]["contentStart"]["role"] == "ASSISTANT"
    assert "textInput" in parsed_events[4]["event"]
    assert parsed_events[4]["event"]["textInput"]["content"] == "Hi there!"

    # Check third message (assistant - last text message)
    assert "contentStart" in parsed_events[6]["event"]
    assert parsed_events[6]["event"]["contentStart"]["role"] == "ASSISTANT"
    assert "textInput" in parsed_events[7]["event"]
    assert parsed_events[7]["event"]["textInput"]["content"] == "The answer is 4"


@pytest.mark.asyncio
async def test_message_history_empty_and_edge_cases(nova_model):
    """Test message history conversion with empty and edge cases."""
    nova_model.connection_id = "test-connection"

    # Test with empty messages
    events = nova_model._get_message_history_events([])
    assert len(events) == 0

    # Test with message containing no text content
    messages = [{"role": "user", "content": []}]
    events = nova_model._get_message_history_events(messages)
    assert len(events) == 0  # No events generated for empty content

    # Test with multiple text blocks in one message
    messages = [{"role": "user", "content": [{"text": "First part"}, {"text": "Second part"}]}]
    events = nova_model._get_message_history_events(messages)
    assert len(events) == 3  # contentStart, textInput, contentEnd
    parsed = json.loads(events[1])
    content = parsed["event"]["textInput"]["content"]
    assert "First part" in content
    assert "Second part" in content


# Error Handling Tests


@pytest.mark.asyncio
async def test_custom_audio_rates_in_events(model_id, boto_session):
    """Test that audio events use configured sample rates."""
    # Create model with custom audio configuration
    provider_config = {"audio": {"output_rate": 48000, "channels": 2}}
    model = BidiNovaSonicModel(
        model_id=model_id, client_config={"boto_session": boto_session}, provider_config=provider_config
    )

    # Test audio output event uses custom configuration
    audio_bytes = b"test audio data"
    audio_base64 = base64.b64encode(audio_bytes).decode("utf-8")
    nova_event = {"audioOutput": {"content": audio_base64}}
    result = model._convert_nova_event(nova_event)

    assert result is not None
    assert isinstance(result, BidiAudioStreamEvent)
    # Should use configured rates, not constants
    assert result.sample_rate == 48000  # Custom config
    assert result.channels == 2  # Custom config
    assert result.format == "pcm"


@pytest.mark.asyncio
async def test_default_audio_rates_in_events(model_id, boto_session):
    """Test that audio events use default sample rates when no custom config."""
    # Create model without custom audio configuration
    model = BidiNovaSonicModel(model_id=model_id, client_config={"boto_session": boto_session})

    # Test audio output event uses defaults
    audio_bytes = b"test audio data"
    audio_base64 = base64.b64encode(audio_bytes).decode("utf-8")
    nova_event = {"audioOutput": {"content": audio_base64}}
    result = model._convert_nova_event(nova_event)

    assert result is not None
    assert isinstance(result, BidiAudioStreamEvent)
    # Should use default rates
    assert result.sample_rate == 16000  # Default output rate
    assert result.channels == 1  # Default channels
    assert result.format == "pcm"


# Nova Sonic v2 Support Tests


def test_nova_sonic_model_constants():
    """Test that Nova Sonic model ID constants are correctly defined."""
    assert NOVA_SONIC_V1_MODEL_ID == "amazon.nova-sonic-v1:0"
    assert NOVA_SONIC_V2_MODEL_ID == "amazon.nova-2-sonic-v1:0"


@pytest.mark.asyncio
async def test_nova_sonic_v1_instantiation(boto_session, mock_client):
    """Test direct instantiation with Nova Sonic v1 model ID."""
    _ = mock_client  # Ensure mock is active

    # Test default creation
    model = BidiNovaSonicModel(model_id=NOVA_SONIC_V1_MODEL_ID, client_config={"boto_session": boto_session})
    assert model.model_id == NOVA_SONIC_V1_MODEL_ID
    assert model.region == "us-east-1"

    # Test with custom config
    provider_config = {"audio": {"voice": "joanna", "output_rate": 24000}}
    client_config = {"boto_session": boto_session}
    model_custom = BidiNovaSonicModel(
        model_id=NOVA_SONIC_V1_MODEL_ID, provider_config=provider_config, client_config=client_config
    )

    assert model_custom.model_id == NOVA_SONIC_V1_MODEL_ID
    assert model_custom.config["audio"]["voice"] == "joanna"
    assert model_custom.config["audio"]["output_rate"] == 24000


@pytest.mark.asyncio
async def test_nova_sonic_v2_instantiation(boto_session, mock_client):
    """Test direct instantiation with Nova Sonic v2 model ID."""
    _ = mock_client  # Ensure mock is active

    # Test default creation
    model = BidiNovaSonicModel(model_id=NOVA_SONIC_V2_MODEL_ID, client_config={"boto_session": boto_session})
    assert model.model_id == NOVA_SONIC_V2_MODEL_ID
    assert model.region == "us-east-1"

    # Test with custom config
    provider_config = {"audio": {"voice": "ruth", "input_rate": 48000}, "inference": {"temperature": 0.8}}
    client_config = {"boto_session": boto_session}
    model_custom = BidiNovaSonicModel(
        model_id=NOVA_SONIC_V2_MODEL_ID, provider_config=provider_config, client_config=client_config
    )

    assert model_custom.model_id == NOVA_SONIC_V2_MODEL_ID
    assert model_custom.config["audio"]["voice"] == "ruth"
    assert model_custom.config["audio"]["input_rate"] == 48000
    assert model_custom.config["inference"]["temperature"] == 0.8


@pytest.mark.asyncio
async def test_nova_sonic_v1_v2_compatibility(boto_session, mock_client):
    """Test that v1 and v2 models have the same config structure and behavior."""
    _ = mock_client  # Ensure mock is active

    # Create both models with same config
    provider_config = {"audio": {"voice": "matthew"}}
    client_config = {"boto_session": boto_session}

    model_v1 = BidiNovaSonicModel(
        model_id=NOVA_SONIC_V1_MODEL_ID, provider_config=provider_config, client_config=client_config
    )
    model_v2 = BidiNovaSonicModel(
        model_id=NOVA_SONIC_V2_MODEL_ID, provider_config=provider_config, client_config=client_config
    )

    # Both should have the same config structure
    assert model_v1.config["audio"]["voice"] == model_v2.config["audio"]["voice"]
    assert model_v1.region == model_v2.region

    # Only model_id should differ
    assert model_v1.model_id != model_v2.model_id
    assert model_v1.model_id == NOVA_SONIC_V1_MODEL_ID
    assert model_v2.model_id == NOVA_SONIC_V2_MODEL_ID


@pytest.mark.asyncio
async def test_backward_compatibility(boto_session, mock_client):
    """Test that existing code continues to work (backward compatibility)."""
    _ = mock_client  # Ensure mock is active

    # Test that default behavior now uses v2 (updated default)
    model_default = BidiNovaSonicModel(client_config={"boto_session": boto_session})
    assert model_default.model_id == NOVA_SONIC_V2_MODEL_ID

    # Test that existing explicit v1 usage still works
    model_explicit_v1 = BidiNovaSonicModel(
        model_id=NOVA_SONIC_V1_MODEL_ID, client_config={"boto_session": boto_session}
    )
    assert model_explicit_v1.model_id == NOVA_SONIC_V1_MODEL_ID

    # Test that explicit v2 usage works
    model_explicit_v2 = BidiNovaSonicModel(
        model_id=NOVA_SONIC_V2_MODEL_ID, client_config={"boto_session": boto_session}
    )
    assert model_explicit_v2.model_id == NOVA_SONIC_V2_MODEL_ID


@pytest.mark.asyncio
async def test_turn_detection_v1_validation(boto_session, mock_client):
    """Test that turn_detection raises error when used with v1 model."""
    _ = mock_client  # Ensure mock is active

    # Test that turn_detection with v1 raises ValueError
    with pytest.raises(ValueError, match="turn_detection is only supported in Nova Sonic v2"):
        BidiNovaSonicModel(
            model_id=NOVA_SONIC_V1_MODEL_ID,
            provider_config={"turn_detection": {"endpointingSensitivity": "MEDIUM"}},
            client_config={"boto_session": boto_session},
        )

    # Test that turn_detection with v2 works fine
    model_v2 = BidiNovaSonicModel(
        model_id=NOVA_SONIC_V2_MODEL_ID,
        provider_config={"turn_detection": {"endpointingSensitivity": "MEDIUM"}},
        client_config={"boto_session": boto_session},
    )
    assert model_v2.config["turn_detection"]["endpointingSensitivity"] == "MEDIUM"

    # Test that empty turn_detection dict doesn't raise error for v1
    model_v1_empty = BidiNovaSonicModel(
        model_id=NOVA_SONIC_V1_MODEL_ID,
        provider_config={"turn_detection": {}},
        client_config={"boto_session": boto_session},
    )
    assert model_v1_empty.model_id == NOVA_SONIC_V1_MODEL_ID


@pytest.mark.asyncio
async def test_turn_detection_sensitivity_validation(boto_session, mock_client):
    """Test that endpointingSensitivity is validated at initialization."""
    _ = mock_client  # Ensure mock is active

    # Test invalid sensitivity value raises ValueError at init
    with pytest.raises(ValueError, match="Invalid endpointingSensitivity.*Must be HIGH, MEDIUM, or LOW"):
        BidiNovaSonicModel(
            model_id=NOVA_SONIC_V2_MODEL_ID,
            provider_config={"turn_detection": {"endpointingSensitivity": "INVALID"}},
            client_config={"boto_session": boto_session},
        )

    # Test valid sensitivity values work
    for sensitivity in ["HIGH", "MEDIUM", "LOW"]:
        model = BidiNovaSonicModel(
            model_id=NOVA_SONIC_V2_MODEL_ID,
            provider_config={"turn_detection": {"endpointingSensitivity": sensitivity}},
            client_config={"boto_session": boto_session},
        )
        assert model.config["turn_detection"]["endpointingSensitivity"] == sensitivity

    # Test that turn_detection without sensitivity works (sensitivity is optional)
    model_no_sensitivity = BidiNovaSonicModel(
        model_id=NOVA_SONIC_V2_MODEL_ID,
        provider_config={"turn_detection": {}},
        client_config={"boto_session": boto_session},
    )
    assert "endpointingSensitivity" not in model_no_sensitivity.config["turn_detection"]


# Error Handling Tests
@pytest.mark.asyncio
async def test_bidi_nova_sonic_model_receive_timeout(nova_model, mock_stream):
    mock_output = AsyncMock()
    mock_output.receive.side_effect = ModelTimeoutException("Connection timeout")
    mock_stream.await_output.return_value = (None, mock_output)

    await nova_model.start()

    with pytest.raises(BidiModelTimeoutError, match=r"Connection timeout"):
        async for _ in nova_model.receive():
            pass


@pytest.mark.asyncio
async def test_bidi_nova_sonic_model_receive_timeout_validation(nova_model, mock_stream):
    mock_output = AsyncMock()
    mock_output.receive.side_effect = ValidationException("InternalErrorCode=531: Request timeout")
    mock_stream.await_output.return_value = (None, mock_output)

    await nova_model.start()

    with pytest.raises(BidiModelTimeoutError, match=r"InternalErrorCode=531"):
        async for _ in nova_model.receive():
            pass


@pytest.mark.asyncio
async def test_error_handling(nova_model, mock_stream):
    """Test error handling in various scenarios."""

    # Test response processor handles errors gracefully
    async def mock_error(*args, **kwargs):
        raise Exception("Test error")

    mock_stream.await_output.side_effect = mock_error

    await nova_model.start()

    # Wait a bit for response processor to handle error
    await asyncio.sleep(0.1)

    # Should still be able to close cleanly
    await nova_model.stop()


# Tool Result Content Tests


@pytest.mark.asyncio
async def test_tool_result_single_content_unwrapped(nova_model, mock_stream):
    """Test that single content item is unwrapped (optimization)."""
    await nova_model.start()

    tool_result: ToolResult = {
        "toolUseId": "tool-123",
        "status": "success",
        "content": [{"text": "Single result"}],
    }

    await nova_model.send(ToolResultEvent(tool_result))

    # Verify events were sent
    assert mock_stream.input_stream.send.called
    calls = mock_stream.input_stream.send.call_args_list

    # Find the toolResult event
    tool_result_events = []
    for call in calls:
        event_json = call.args[0].value.bytes_.decode("utf-8")
        event = json.loads(event_json)
        if "toolResult" in event.get("event", {}):
            tool_result_events.append(event)

    assert len(tool_result_events) > 0
    tool_result_event = tool_result_events[0]["event"]["toolResult"]

    # Single content should be unwrapped (not in array)
    content = json.loads(tool_result_event["content"])
    assert content == {"text": "Single result"}

    await nova_model.stop()


@pytest.mark.asyncio
async def test_tool_result_multiple_content_as_array(nova_model, mock_stream):
    """Test that multiple content items are sent as array."""
    await nova_model.start()

    tool_result: ToolResult = {
        "toolUseId": "tool-456",
        "status": "success",
        "content": [{"text": "Part 1"}, {"json": {"data": "value"}}],
    }

    await nova_model.send(ToolResultEvent(tool_result))

    # Verify events were sent
    assert mock_stream.input_stream.send.called
    calls = mock_stream.input_stream.send.call_args_list

    # Find the toolResult event
    tool_result_events = []
    for call in calls:
        event_json = call.args[0].value.bytes_.decode("utf-8")
        event = json.loads(event_json)
        if "toolResult" in event.get("event", {}):
            tool_result_events.append(event)

    assert len(tool_result_events) > 0
    tool_result_event = tool_result_events[0]["event"]["toolResult"]

    # Multiple content should be in array format
    content = json.loads(tool_result_event["content"])
    assert "content" in content
    assert isinstance(content["content"], list)
    assert len(content["content"]) == 2
    assert content["content"][0] == {"text": "Part 1"}
    assert content["content"][1] == {"json": {"data": "value"}}

    await nova_model.stop()


@pytest.mark.asyncio
async def test_tool_result_empty_content(nova_model, mock_stream):
    """Test that empty content is handled gracefully."""
    await nova_model.start()

    tool_result: ToolResult = {
        "toolUseId": "tool-789",
        "status": "success",
        "content": [],
    }

    await nova_model.send(ToolResultEvent(tool_result))

    # Verify events were sent
    assert mock_stream.input_stream.send.called
    calls = mock_stream.input_stream.send.call_args_list

    # Find the toolResult event
    tool_result_events = []
    for call in calls:
        event_json = call.args[0].value.bytes_.decode("utf-8")
        event = json.loads(event_json)
        if "toolResult" in event.get("event", {}):
            tool_result_events.append(event)

    assert len(tool_result_events) > 0
    tool_result_event = tool_result_events[0]["event"]["toolResult"]

    # Empty content should result in empty array wrapped in content key
    content = json.loads(tool_result_event["content"])
    assert content == {"content": []}

    await nova_model.stop()


@pytest.mark.asyncio
async def test_tool_result_unsupported_content_type(nova_model):
    """Test that unsupported content types raise ValueError."""
    await nova_model.start()

    # Test with image content (unsupported)
    tool_result_image: ToolResult = {
        "toolUseId": "tool-999",
        "status": "success",
        "content": [{"image": {"format": "jpeg", "source": {"bytes": b"image_data"}}}],
    }

    with pytest.raises(ValueError, match=r"Content type not supported by Nova Sonic"):
        await nova_model.send(ToolResultEvent(tool_result_image))

    # Test with document content (unsupported)
    tool_result_doc: ToolResult = {
        "toolUseId": "tool-888",
        "status": "success",
        "content": [{"document": {"format": "pdf", "source": {"bytes": b"doc_data"}}}],
    }

    with pytest.raises(ValueError, match=r"Content type not supported by Nova Sonic"):
        await nova_model.send(ToolResultEvent(tool_result_doc))

    # Test with mixed content (one unsupported)
    tool_result_mixed: ToolResult = {
        "toolUseId": "tool-777",
        "status": "success",
        "content": [{"text": "Valid text"}, {"image": {"format": "jpeg", "source": {"bytes": b"image_data"}}}],
    }

    with pytest.raises(ValueError, match=r"Content type not supported by Nova Sonic"):
        await nova_model.send(ToolResultEvent(tool_result_mixed))

    await nova_model.stop()



"""Bidirectional streaming package."""

from typing import Any

# Main components - Primary user interface
# Re-export standard agent events for tool handling
from ...types._events import (
    ToolResultEvent,
    ToolStreamEvent,
    ToolUseStreamEvent,
)
from .agent.agent import BidiAgent

# Model interface (for custom implementations)
from .models.model import BidiModel

# Built-in tools
from .tools import stop_conversation

# Event types - For type hints and event handling
from .types.events import (
    BidiAudioInputEvent,
    BidiAudioStreamEvent,
    BidiConnectionCloseEvent,
    BidiConnectionStartEvent,
    BidiErrorEvent,
    BidiImageInputEvent,
    BidiInputEvent,
    BidiInterruptionEvent,
    BidiOutputEvent,
    BidiResponseCompleteEvent,
    BidiResponseStartEvent,
    BidiTextInputEvent,
    BidiTranscriptStreamEvent,
    BidiUsageEvent,
    ModalityUsage,
)

__all__ = [
    # Main interface
    "BidiAgent",
    # Built-in tools
    "stop_conversation",
    # Input Event types
    "BidiTextInputEvent",
    "BidiAudioInputEvent",
    "BidiImageInputEvent",
    "BidiInputEvent",
    # Output Event types
    "BidiConnectionStartEvent",
    "BidiConnectionCloseEvent",
    "BidiResponseStartEvent",
    "BidiResponseCompleteEvent",
    "BidiAudioStreamEvent",
    "BidiTranscriptStreamEvent",
    "BidiInterruptionEvent",
    "BidiUsageEvent",
    "ModalityUsage",
    "BidiErrorEvent",
    "BidiOutputEvent",
    # Tool Event types (reused from standard agent)
    "ToolUseStreamEvent",
    "ToolResultEvent",
    "ToolStreamEvent",
    # Model interface
    "BidiModel",
]


def __getattr__(name: str) -> Any:
    """Lazy load IO implementations only when accessed.

    This defers the import of optional dependencies until actually needed.
    """
    if name == "BidiAudioIO":
        from .io.audio import BidiAudioIO

        return BidiAudioIO
    if name == "BidiTextIO":
        from .io.text import BidiTextIO

        return BidiTextIO
    raise AttributeError(f"cannot import name '{name}' from '{__name__}' ({__file__})")



"""Human-in-the-loop interrupt system for agent workflows."""

from dataclasses import asdict, dataclass, field
from typing import TYPE_CHECKING, Any, cast

if TYPE_CHECKING:
    from .types.agent import AgentInput
    from .types.interrupt import InterruptResponseContent


@dataclass
class Interrupt:
    """Represents an interrupt that can pause agent execution for human-in-the-loop workflows.

    Attributes:
        id: Unique identifier.
        name: User defined name.
        reason: User provided reason for raising the interrupt.
        response: Human response provided when resuming the agent after an interrupt.
    """

    id: str
    name: str
    reason: Any = None
    response: Any = None

    def to_dict(self) -> dict[str, Any]:
        """Serialize to dict for session management."""
        return asdict(self)


class InterruptException(Exception):
    """Exception raised when human input is required."""

    def __init__(self, interrupt: Interrupt) -> None:
        """Set the interrupt."""
        self.interrupt = interrupt


@dataclass
class _InterruptState:
    """Track the state of interrupt events raised by the user.

    Note, interrupt state is cleared after resuming.

    Attributes:
        interrupts: Interrupts raised by the user.
        context: Additional context associated with an interrupt event.
        activated: True if agent is in an interrupt state, False otherwise.
    """

    interrupts: dict[str, Interrupt] = field(default_factory=dict)
    context: dict[str, Any] = field(default_factory=dict)
    activated: bool = False
    _version: int = field(default=0, compare=False, repr=False)

    def activate(self) -> None:
        """Activate the interrupt state."""
        self.activated = True
        self._version += 1

    def deactivate(self) -> None:
        """Deacitvate the interrupt state.

        Interrupts and context are cleared.
        """
        self.interrupts = {}
        self.context = {}
        self.activated = False
        self._version += 1

    def resume(self, prompt: "AgentInput") -> None:
        """Configure the interrupt state if resuming from an interrupt event.

        Args:
            prompt: User responses if resuming from interrupt.

        Raises:
            TypeError: If in interrupt state but user did not provide responses.
        """
        if not self.activated:
            return

        if not isinstance(prompt, list):
            raise TypeError(f"prompt_type={type(prompt)} | must resume from interrupt with list of interruptResponse's")

        invalid_types = [
            content_type for content in prompt for content_type in content if content_type != "interruptResponse"
        ]
        if invalid_types:
            raise TypeError(
                f"content_types=<{invalid_types}> | must resume from interrupt with list of interruptResponse's"
            )

        contents = cast(list["InterruptResponseContent"], prompt)
        for content in contents:
            interrupt_id = content["interruptResponse"]["interruptId"]
            interrupt_response = content["interruptResponse"]["response"]

            if interrupt_id not in self.interrupts:
                raise KeyError(f"interrupt_id=<{interrupt_id}> | no interrupt found")

            self.interrupts[interrupt_id].response = interrupt_response

        self.context["responses"] = contents
        self._version += 1

    def _get_version(self) -> int:
        """Get the current version number of the interrupt state.

        The version is incremented each time activate(), deactivate(), or resume() is called.
        Consumers can compare versions to detect changes without requiring
        explicit dirty flag clearing.

        Returns:
            The current version number.
        """
        return self._version

    def to_dict(self) -> dict[str, Any]:
        """Serialize to dict for session management."""
        return {
            "interrupts": {k: v.to_dict() for k, v in self.interrupts.items()},
            "context": self.context,
            "activated": self.activated,
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "_InterruptState":
        """Initiailize interrupt state from serialized interrupt state.

        Interrupt state can be serialized with the `to_dict` method.
        """
        return cls(
            interrupts={
                interrupt_id: Interrupt(**interrupt_data) for interrupt_id, interrupt_data in data["interrupts"].items()
            },
            context=data["context"],
            activated=data["activated"],
        )



"""Unit tests for retry strategy implementations."""

from unittest.mock import Mock

import pytest

from strands import ModelRetryStrategy
from strands.hooks import AfterInvocationEvent, AfterModelCallEvent, HookRegistry
from strands.types._events import EventLoopThrottleEvent
from strands.types.exceptions import ModelThrottledException

# ModelRetryStrategy Tests


def test_model_retry_strategy_init_with_defaults():
    """Test ModelRetryStrategy initialization with default parameters."""
    strategy = ModelRetryStrategy()
    assert strategy._max_attempts == 6
    assert strategy._initial_delay == 4
    assert strategy._max_delay == 240
    assert strategy._current_attempt == 0


def test_model_retry_strategy_init_with_custom_parameters():
    """Test ModelRetryStrategy initialization with custom parameters."""
    strategy = ModelRetryStrategy(max_attempts=3, initial_delay=2, max_delay=60)
    assert strategy._max_attempts == 3
    assert strategy._initial_delay == 2
    assert strategy._max_delay == 60
    assert strategy._current_attempt == 0


def test_model_retry_strategy_calculate_delay_with_different_attempts():
    """Test _calculate_delay returns correct exponential backoff for different attempt numbers."""
    strategy = ModelRetryStrategy(initial_delay=2, max_delay=32)

    # Test exponential backoff: 2 * (2^attempt)
    assert strategy._calculate_delay(0) == 2  # 2 * 2^0 = 2
    assert strategy._calculate_delay(1) == 4  # 2 * 2^1 = 4
    assert strategy._calculate_delay(2) == 8  # 2 * 2^2 = 8
    assert strategy._calculate_delay(3) == 16  # 2 * 2^3 = 16
    assert strategy._calculate_delay(4) == 32  # 2 * 2^4 = 32 (at max)
    assert strategy._calculate_delay(5) == 32  # 2 * 2^5 = 64, capped at 32
    assert strategy._calculate_delay(10) == 32  # Large attempt, still capped


def test_model_retry_strategy_calculate_delay_respects_max_delay():
    """Test _calculate_delay respects max_delay cap."""
    strategy = ModelRetryStrategy(initial_delay=10, max_delay=50)

    assert strategy._calculate_delay(0) == 10  # 10 * 2^0 = 10
    assert strategy._calculate_delay(1) == 20  # 10 * 2^1 = 20
    assert strategy._calculate_delay(2) == 40  # 10 * 2^2 = 40
    assert strategy._calculate_delay(3) == 50  # 10 * 2^3 = 80, capped at 50
    assert strategy._calculate_delay(4) == 50  # 10 * 2^4 = 160, capped at 50


def test_model_retry_strategy_register_hooks():
    """Test that ModelRetryStrategy registers AfterModelCallEvent and AfterInvocationEvent callbacks."""
    strategy = ModelRetryStrategy()
    registry = HookRegistry()

    strategy.register_hooks(registry)

    # Verify AfterModelCallEvent callback was registered
    assert AfterModelCallEvent in registry._registered_callbacks
    assert len(registry._registered_callbacks[AfterModelCallEvent]) == 1

    # Verify AfterInvocationEvent callback was registered
    assert AfterInvocationEvent in registry._registered_callbacks
    assert len(registry._registered_callbacks[AfterInvocationEvent]) == 1


@pytest.mark.asyncio
async def test_model_retry_strategy_retry_on_throttle_exception_first_attempt(mock_sleep):
    """Test retry behavior on first ModelThrottledException."""
    strategy = ModelRetryStrategy(max_attempts=3, initial_delay=2, max_delay=60)
    mock_agent = Mock()

    event = AfterModelCallEvent(
        agent=mock_agent,
        exception=ModelThrottledException("Throttled"),
    )

    await strategy._handle_after_model_call(event)

    # Should set retry to True
    assert event.retry is True
    # Should sleep for initial_delay (attempt 0: 2 * 2^0 = 2)
    assert mock_sleep.sleep_calls == [2]
    assert mock_sleep.sleep_calls[0] == strategy._calculate_delay(0)
    # Should increment attempt
    assert strategy._current_attempt == 1


@pytest.mark.asyncio
async def test_model_retry_strategy_exponential_backoff(mock_sleep):
    """Test exponential backoff calculation."""
    strategy = ModelRetryStrategy(max_attempts=5, initial_delay=2, max_delay=16)
    mock_agent = Mock()

    # Simulate multiple retries
    for _ in range(4):
        event = AfterModelCallEvent(
            agent=mock_agent,
            exception=ModelThrottledException("Throttled"),
        )
        await strategy._handle_after_model_call(event)
        assert event.retry is True

    # Verify exponential backoff with max_delay cap
    # attempt 0: 2*2^0=2, attempt 1: 2*2^1=4, attempt 2: 2*2^2=8, attempt 3: 2*2^3=16 (capped)
    assert mock_sleep.sleep_calls == [2, 4, 8, 16]
    for i, sleep_delay in enumerate(mock_sleep.sleep_calls):
        assert sleep_delay == strategy._calculate_delay(i)


@pytest.mark.asyncio
async def test_model_retry_strategy_no_retry_after_max_attempts(mock_sleep):
    """Test that retry is not set after reaching max_attempts."""
    strategy = ModelRetryStrategy(max_attempts=2, initial_delay=2, max_delay=60)
    mock_agent = Mock()

    # First attempt
    event1 = AfterModelCallEvent(
        agent=mock_agent,
        exception=ModelThrottledException("Throttled"),
    )
    await strategy._handle_after_model_call(event1)
    assert event1.retry is True
    assert strategy._current_attempt == 1

    # Second attempt (at max_attempts)
    event2 = AfterModelCallEvent(
        agent=mock_agent,
        exception=ModelThrottledException("Throttled"),
    )
    await strategy._handle_after_model_call(event2)
    # Should NOT retry after reaching max_attempts
    assert event2.retry is False
    assert strategy._current_attempt == 2


@pytest.mark.asyncio
async def test_model_retry_strategy_no_retry_on_non_throttle_exception():
    """Test that retry is not set for non-throttling exceptions."""
    strategy = ModelRetryStrategy()
    mock_agent = Mock()

    event = AfterModelCallEvent(
        agent=mock_agent,
        exception=ValueError("Some other error"),
    )

    await strategy._handle_after_model_call(event)

    # Should not retry on non-throttling exceptions
    assert event.retry is False
    assert strategy._current_attempt == 0


@pytest.mark.asyncio
async def test_model_retry_strategy_no_retry_on_success():
    """Test that retry is not set when model call succeeds."""
    strategy = ModelRetryStrategy()
    mock_agent = Mock()

    event = AfterModelCallEvent(
        agent=mock_agent,
        stop_response=AfterModelCallEvent.ModelStopResponse(
            message={"role": "assistant", "content": [{"text": "Success"}]},
            stop_reason="end_turn",
        ),
    )

    await strategy._handle_after_model_call(event)

    # Should not retry on success
    assert event.retry is False


@pytest.mark.asyncio
async def test_model_retry_strategy_reset_on_success(mock_sleep):
    """Test that strategy resets attempt counter on successful call."""
    strategy = ModelRetryStrategy(max_attempts=3, initial_delay=2, max_delay=60)
    mock_agent = Mock()

    # First failure
    event1 = AfterModelCallEvent(
        agent=mock_agent,
        exception=ModelThrottledException("Throttled"),
    )
    await strategy._handle_after_model_call(event1)
    assert event1.retry is True
    assert strategy._current_attempt == 1
    # Should sleep for initial_delay (attempt 0: 2 * 2^0 = 2)
    assert mock_sleep.sleep_calls == [2]
    assert mock_sleep.sleep_calls[0] == strategy._calculate_delay(0)

    # Success - should reset
    event2 = AfterModelCallEvent(
        agent=mock_agent,
        stop_response=AfterModelCallEvent.ModelStopResponse(
            message={"role": "assistant", "content": [{"text": "Success"}]},
            stop_reason="end_turn",
        ),
    )
    await strategy._handle_after_model_call(event2)
    assert event2.retry is False
    # Should reset to initial state
    assert strategy._current_attempt == 0
    assert strategy._calculate_delay(0) == 2


@pytest.mark.asyncio
async def test_model_retry_strategy_skips_if_already_retrying():
    """Test that strategy skips processing if event.retry is already True."""
    strategy = ModelRetryStrategy(max_attempts=3, initial_delay=2, max_delay=60)
    mock_agent = Mock()

    event = AfterModelCallEvent(
        agent=mock_agent,
        exception=ModelThrottledException("Throttled"),
    )
    # Simulate another hook already set retry to True
    event.retry = True

    await strategy._handle_after_model_call(event)

    # Should not modify state since another hook already triggered retry
    assert strategy._current_attempt == 0
    assert event.retry is True


@pytest.mark.asyncio
async def test_model_retry_strategy_reset_on_after_invocation():
    """Test that strategy resets state on AfterInvocationEvent."""
    strategy = ModelRetryStrategy(max_attempts=3, initial_delay=2, max_delay=60)
    mock_agent = Mock()

    # Simulate some retry attempts
    strategy._current_attempt = 3

    event = AfterInvocationEvent(agent=mock_agent, result=Mock())
    await strategy._handle_after_invocation(event)

    # Should reset to initial state
    assert strategy._current_attempt == 0


@pytest.mark.asyncio
async def test_model_retry_strategy_backwards_compatible_event_set_on_retry(mock_sleep):
    """Test that _backwards_compatible_event_to_yield is set when retrying."""
    strategy = ModelRetryStrategy(max_attempts=3, initial_delay=2, max_delay=60)
    mock_agent = Mock()

    event = AfterModelCallEvent(
        agent=mock_agent,
        exception=ModelThrottledException("Throttled"),
    )

    await strategy._handle_after_model_call(event)

    # Should have set the backwards compatible event
    assert strategy._backwards_compatible_event_to_yield is not None
    assert isinstance(strategy._backwards_compatible_event_to_yield, EventLoopThrottleEvent)
    assert strategy._backwards_compatible_event_to_yield["event_loop_throttled_delay"] == 2


@pytest.mark.asyncio
async def test_model_retry_strategy_backwards_compatible_event_cleared_on_success():
    """Test that _backwards_compatible_event_to_yield is cleared on success."""
    strategy = ModelRetryStrategy(max_attempts=3, initial_delay=2, max_delay=60)
    mock_agent = Mock()

    # Set a previous backwards compatible event
    strategy._backwards_compatible_event_to_yield = EventLoopThrottleEvent(delay=2)

    event = AfterModelCallEvent(
        agent=mock_agent,
        stop_response=AfterModelCallEvent.ModelStopResponse(
            message={"role": "assistant", "content": [{"text": "Success"}]},
            stop_reason="end_turn",
        ),
    )

    await strategy._handle_after_model_call(event)

    # Should have cleared the backwards compatible event
    assert strategy._backwards_compatible_event_to_yield is None


@pytest.mark.asyncio
async def test_model_retry_strategy_backwards_compatible_event_not_set_on_max_attempts(mock_sleep):
    """Test that _backwards_compatible_event_to_yield is not set when max attempts reached."""
    strategy = ModelRetryStrategy(max_attempts=1, initial_delay=2, max_delay=60)
    mock_agent = Mock()

    event = AfterModelCallEvent(
        agent=mock_agent,
        exception=ModelThrottledException("Throttled"),
    )

    await strategy._handle_after_model_call(event)

    # Should not have set the backwards compatible event since max attempts reached
    assert strategy._backwards_compatible_event_to_yield is None
    assert event.retry is False


@pytest.mark.asyncio
async def test_model_retry_strategy_no_retry_when_no_exception_and_no_stop_response():
    """Test that retry is not set when there's no exception and no stop_response."""
    strategy = ModelRetryStrategy()
    mock_agent = Mock()

    # Event with neither exception nor stop_response
    event = AfterModelCallEvent(
        agent=mock_agent,
        exception=None,
        stop_response=None,
    )

    await strategy._handle_after_model_call(event)

    # Should not retry and should reset state
    assert event.retry is False
    assert strategy._current_attempt == 0
